r"""
This module provides a command-line interface for running various reference tissue model (RTM) analyses.

The module defines a set of command-line arguments for the input and output data, as well
as the analysis specifics. Several types of RTM analyses, such as SRTM, FRTM, MRTM and its variants,
are supported, each with its associated arguments.

The primary function in the module is `main()`, which is responsible for parsing command-line
arguments, running the selected analysis, and handling the results. The analysis is performed
using the :class:`RTMAnalysis<petpal.reference_tissue_models.RTMAnalysis>` class.

Example:
    Assuming that the package has been installed, we have:
    
    For running an FRTM analysis:
    
    .. code-block:: bash
    
        petpal-rtms frtm --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --initial-guesses 0.1 0.1 0.1 0.1 0.1 --lower-bounds 0.0 0.0 0.0 0.0 0.0 --upper-bounds 5.0 5.0 5.0 5.0 5.0
    
    For running an SRTM analysis:
    
    .. code-block:: bash
    
        petpal-rtms srtm --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --initial-guesses 0.1 0.1 0.1 0.1 0.1 --lower-bounds 0.0 0.0 0.0 0.0 0.0 --upper-bounds 5.0 5.0 5.0 5.0 5.0
        
    For running an FRTM2 analysis:
    
    .. code-block:: bash
    
        petpal-rtms frtm2 --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --k2-prime 0.5 --prefix sub_001 --print --initial-guesses 0.1 0.1 0.1 --lower-bounds 0.0 0.0 0.0 --upper-bounds 5.0 5.0 5.0
    
    For running an SRTM2 analysis:
    
    .. code-block:: bash
    
        petpal-rtms srtm2 --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --k2-prime 0.5 --prefix sub_001 --print --initial-guesses 0.1 0.1 --lower-bounds 0.0 0.0 0.0 --upper-bounds 5.0 5.0
    
    For running an MRTM analysis:
    
    .. code-block:: bash
    
        petpal-rtms mrtm --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --t-thresh-in-mins 30.0
    
    For running an MRTM2 analysis:
    
    .. code-block:: bash
    
        petpal-rtms mrtm2 --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --t-thresh-in-mins 30.0 --k2-prime 0.1
    
    
    For running the original MRTM analysis:
    
    .. code-block:: bash
    
        petpal-rtms mrtm-original --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --t-thresh-in-mins 30.0
    
    
    

See Also:
    * :mod:`petpal.reference_tissue_models`
    
"""

import argparse
import numpy as np
from typing import Union
from ..kinetic_modeling.rtm_analysis import RTMAnalysis


_RTM_EXAMPLES_ = (r"""
Examples:
  - FRTM (4 parameters: R1, k2, k3, k4):
    petpal-rtms frtm --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --initial-guesses 0.1 0.1 0.1 0.1 --lower-bounds 0.0 0.0 0.0 0.0 --upper-bounds 5.0 5.0 5.0 5.0
  - SRTM (3 parameters: R1, k2, BP):
    petpal-rtms srtm --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --initial-guesses 0.1 0.1 0.1 --lower-bounds 0.0 0.0 0.0 --upper-bounds 5.0 5.0 5.0
  - FRTM2 (3 parameters: R1, k3, k4):
    petpal-rtms frtm2 --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --k2-prime 0.5 --prefix sub_001 --print --initial-guesses 0.1 0.1 0.1 0.1 --lower-bounds 0.0 0.0 0.0 0.0 --upper-bounds 5.0 5.0 5.0 5.0
  - SRTM2 (2 parameters: R1, BP):
    petpal-rtms srtm2 --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --k2-prime 0.5 --prefix sub_001 --print --initial-guesses 0.1 0.1 0.1 0.1 --lower-bounds 0.0 0.0 0.0 --upper-bounds 5.0 5.0
  - MRTM (2 parameters: BP, k2_prime):
    petpal-rtms mrtm --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --t-thresh-in-mins 30.0
  - MRTM2 (2 parameters: BP):
    petpal-rtms mrtm2 --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --t-thresh-in-mins 30.0 --k2-prime 0.1
  - MRTM (Original) (2 parameters: BP, k2_prime):
    petpal-rtms mrtm-original --ref-tac-path /path/to/ref/tac --roi-tac-path /path/to/roi/tac --prefix sub_001 --print --t-thresh-in-mins 30.0
""")


def add_common_args(parser):
    r"""
    Adds common arguments to an ArgumentParser object.

    This function modifies the passed ArgumentParser object by adding five arguments commonly used for the command line
    scripts. It uses the add_argument method of the ArgumentParser class. After running this function, the parser will
    be able to accept and parse these additional arguments from the command line when run.

    Args:
        parser (argparse.ArgumentParser): The argument parser object to which the arguments will be added.

    Raises:
        argparse.ArgumentError: If a duplicate argument is added.

    Side Effects:
        The ArgumentParser object is modified by adding new arguments.
    """
    parser.add_argument('-e', '--ref-tac-path', required=True, type=str,
                        help='Absolute path for reference TAC')
    parser.add_argument('-i', '--roi-tac-path', required=True, type=str,
                        help='Absolute path for ROI TAC')
    parser.add_argument('-o', '--output-directory', required=True, type=str,
                        help='Absolute path for the output directory')
    parser.add_argument('-f', '--prefix', required=True, type=str,
                        help='Prefix for the output filename of the result')
    parser.add_argument('--print', required=False, action='store_true',
                        help="Whether to print the fitting results to screen.")


def parse_args():
    """
    Defines and parses command-line arguments.

    This function creates an ArgumentParser object, defines command-line arguments using that object,
    and then calls its parse_args method and returns the results.

    The function defines a set of sub-commands (srtm, frtm, mrtm, mrtm2, mrtm-original), each of which has a different
    set of requiring arguments. Each sub-command represents a different analysis method and requires different inputs.

    Returns:
        argparse.Namespace: An object representing the command-line arguments. The object's attributes are the
        various command-line arguments, where the attribute name matches the option string.

    Raises:
        argparse.ArgumentError: If the argument is already defined.
        argparse.ArgumentTypeError: If the argument type conversion function raises this exception.
        SystemExit: If invalid arguments are passed or `-h`/`--help` is chosen.

    """
    parser = argparse.ArgumentParser(prog='petpal-rtms',
                                     description='Command line interface for running RTM analyses on TACs',
                                     epilog=_RTM_EXAMPLES_,
                                     formatter_class=argparse.RawTextHelpFormatter)
    subparsers = parser.add_subparsers(dest="method", help="Analysis method")

    parser_srtm = subparsers.add_parser('srtm', help='Perform SRTM analysis')
    parser_frtm = subparsers.add_parser('frtm', help='Perform FRTM analysis')
    parser_srtm_2 = subparsers.add_parser('srtm2', help='Perform SRTM2 analysis')
    parser_frtm_2 = subparsers.add_parser('frtm2', help='Perform FRTM2 analysis')
    
    parser_mrtm = subparsers.add_parser('mrtm',
                                        help='Perform Ichise\'s MRTM (2003) analysis')
    parser_mrtm_2 = subparsers.add_parser('mrtm2',
                                          help='Perform Ichise\'s MRTM2 (2003) analysis')
    parser_mrtm_original = subparsers.add_parser('mrtm-original',
                                                 help='Perform Ichise\'s original MRTM (1996) analysis')
    parser_list = [parser_srtm, parser_frtm, parser_srtm_2, parser_frtm_2,
                   parser_mrtm, parser_mrtm_2, parser_mrtm_original]
    
    for a_parser in parser_list[:]:
        add_common_args(a_parser)
        if a_parser.prog.endswith('2'):
            a_parser.add_argument('-k', '--k2-prime', required=True, type=float,
                                  help='k2_prime value for the reduced RTM analysis.')
            
        if 'mrtm' in a_parser.prog:
            a_parser.add_argument('-t', '--t-thresh-in-mins', required=True, type=float,
                                  help='Threshold time in minutes for the MRTM analyses.')
        else:
            a_parser.add_argument("-g", "--initial-guesses", required=False, nargs='+', type=float,
                                      help="Initial guesses for each fitting parameter.")
            a_parser.add_argument("-l", "--lower-bounds", required=False, nargs='+', type=float,
                                      help="Lower bounds for each fitting parameter.")
            a_parser.add_argument("-u", "--upper-bounds", required=False, nargs='+', type=float,
                                      help="Upper bounds for each fitting parameter.")

    return parser.parse_args()


def _generate_bounds(initial: Union[list, None],
                     lower: Union[list, None],
                     upper: Union[list, None]) -> Union[np.ndarray, None]:
    r"""
    Generates the bounds for the fitting parameters.

    This function takes lists of initial fitting parameters, lower bounds, and upper bounds.
    All lists must have the same length. If no initial parameters are provided, the function returns None.

    Args:
        initial (list, optional): List of initial guesses for fitting parameters. If None, no bounds are generated.
        lower (list, optional): List of lower bounds for fitting parameters.
        upper (list, optional): List of upper bounds for fitting parameters.

    Returns:
        (np.ndarray, optional): If initial is not None, then we return a numpy array of shape [n, 3],
        where n is the number of parameters, where column 0 has the initial guesses, column 1 has lower bounds, and
        column 2 has upper bounds. If initial is None, function will return None.

    Raises:
        ValueError: If initial is not None and the length of initial, lower, and upper are not the same.
    """
    if initial is not None:
        if upper is None or lower is None:
            raise ValueError("The lower and upper bounds must also be set.")
        if (len(initial) != len(lower)) or (len(initial) != len(upper)) or (len(upper) != len(lower)):
            raise ValueError("The number of initial guesses, lower bounds and upper bounds must be the same.")
        return np.asarray(np.asarray([initial, lower, upper]).T)
    else:
        return None


def _pretty_print_results(vals: dict, errs: dict) -> None:
    r"""
    Formats and prints the analysis results to the console.

    This helper function creates a pretty, human-readable formatted string of the parameter names,
    their fitted values, standard errors, and percentage errors, and then prints this
    formatted string to the console.

    Args:
        vals (dict): A dictionary where the keys are the parameter names and the values are fitted values.
        errs (dict): A dictionary where the keys are the parameter names and the values are standard errors.

    Returns:
        None

    Side Effects:
        Prints the formatted result string to the console.

    Raises:
        KeyError: If the keys of 'vals' and 'errs' do not match.
    """
    title_str = f"{'Param':<5} {'FitVal':<6}    {'StdErr':<6} ({'%Err':>6})|"
    print("-" * len(title_str))
    print(title_str)
    print("-" * len(title_str))
    for param_name in vals:
        val = vals[param_name]
        err = errs[param_name]
        print(f"{param_name:<5} {val:<6.4f} +- {err:<6.4f} ({err / val * 100:>5.2f}%)|")
    print("-" * len(title_str))


def main():
    args = parse_args()
    
    if args.method is None:
        args.print_help()
        raise Exception('Exiting without command')
    
    analysis = RTMAnalysis(ref_tac_path=args.ref_tac_path,
                           roi_tac_path=args.roi_tac_path,
                           output_directory=args.output_directory,
                           output_filename_prefix=args.prefix,
                           method=args.method)
    
    if args.method.startswith('mrtm'):
        analysis.run_analysis(t_thresh_in_mins=args.t_thresh_in_mins,
                              k2_prime=getattr(args, 'k2_prime', None))
    else:
        bounds = _generate_bounds(initial=args.initial_guesses, lower=args.lower_bounds, upper=args.upper_bounds)
        if args.method.endswith('2'):
            analysis.run_analysis(bounds=bounds, k2_prime=getattr(args, 'k2_prime', None))
        else:
            analysis.run_analysis(bounds=bounds)
    
    analysis.save_analysis()

    if args.print:
        if args.method.startswith('mrtm'):
            print(f"BP      : {analysis.analysis_props['BP']:<.7f}")
            print(f"k2_prime: {analysis.analysis_props['k2Prime']:<.7f}")
        else:
            _pretty_print_results(vals=analysis.analysis_props["FitValues"],
                                  errs=analysis.analysis_props["FitStdErr"])
    

if __name__ == '__main__':
    main()


"""
Module for functions calculating standard uptake value (SUV) and related measures, such as standard
uptake value ratio (SUVR).
"""
import ants

from ..utils.dimension import gen_3d_img_from_timeseries

from ..utils.stats import mean_value_in_region
from ..utils.math_lib import weighted_sum_computation
from ..utils.useful_functions import nearest_frame_to_timepoint
from ..utils.image_io import (get_half_life_from_nifti,
                              load_metadata_for_nifti_with_same_filename,
                              safe_copy_meta)


def weighted_sum_for_suv(input_image_path: str,
                         output_image_path: str | None,
                         start_time: float=0,
                         end_time: float=-1) -> ants.ANTsImage:
    """Function that calculates the weighted series sum for a PET image specifically for
    calculating the standard uptake value (SUV) of the image.
    
    Args:
        input_image_path (str): Path to a 4D PET image which we calculate the sum on.
        output_image_path (str): Path to which output image is saved. If None, returns
            calculated image without saving.
        start_time: Time in seconds from the start of the scan from which to begin sum calculation.
            Only frames after selected time will be included in the sum. Default 0.
        end_time: Time in seconds from the start of the scan from which to end sum calculation.
            Only frames before selected time will be included in the sum. If -1, use all frames
            after `start_time` in the calculation. Default -1.
            
    Returns:
        weighted_sum_img (ants.ANTsImage): 3D image resulting from the sum calculation.
    """
    half_life = get_half_life_from_nifti(image_path=input_image_path)
    if half_life <= 0:
        raise ValueError('(ImageOps4d): Radioisotope half life is zero or negative.')
    pet_meta = load_metadata_for_nifti_with_same_filename(input_image_path)
    pet_img = ants.image_read(input_image_path)
    frame_start = pet_meta['FrameTimesStart']
    frame_duration = pet_meta['FrameDuration']

    if 'DecayCorrectionFactor' in pet_meta.keys():
        decay_correction = pet_meta['DecayCorrectionFactor']
    elif 'DecayFactor' in pet_meta.keys():
        decay_correction = pet_meta['DecayFactor']
    else:
        raise ValueError("Neither 'DecayCorrectionFactor' nor 'DecayFactor' exist in meta-data "
                         "file")

    last_frame_time = frame_start[-1]
    if end_time!=-1:
        last_frame_time = end_time
    scan_start = frame_start[0]
    nearest_frame = nearest_frame_to_timepoint(frame_times=frame_start)
    calc_first_frame = int(nearest_frame(start_time+scan_start))
    calc_last_frame = int(nearest_frame(last_frame_time+scan_start))
    if calc_first_frame==calc_last_frame:
        calc_last_frame += 1
    pet_series_adjusted = pet_img[:,:,:,calc_first_frame:calc_last_frame]
    frame_start_adjusted = frame_start[calc_first_frame:calc_last_frame]
    frame_duration_adjusted = frame_duration[calc_first_frame:calc_last_frame]
    decay_correction_adjusted = decay_correction[calc_first_frame:calc_last_frame]

    weighted_sum_arr = weighted_sum_computation(frame_duration=frame_duration_adjusted,
                                                half_life=half_life,
                                                pet_series=pet_series_adjusted,
                                                frame_start=frame_start_adjusted,
                                                decay_correction=decay_correction_adjusted)
    weighted_sum_img = ants.from_numpy_like(weighted_sum_arr,gen_3d_img_from_timeseries(pet_img))

    if output_image_path is not None:
        ants.image_write(weighted_sum_img, output_image_path)
        safe_copy_meta(input_image_path=input_image_path,
                       out_image_path=output_image_path)

    return weighted_sum_img


def suv(input_image_path: str,
        output_image_path: str | None,
        weight: float,
        dose: float,
        start_time: float,
        end_time: float) -> ants.ANTsImage:
    """Compute standard uptake value (SUV) over a pet image. Calculate the weighted image sum
    then divide by the dose and multiplying by the weight of the participant.

    Args:
        input_image_path: Path to input PET image.
        output_image_path: Path to which SUV image is saved.
        weight: Weight of the participant in kg.
        dose: Dose injected during the scan in MBq.
        start_time: Start time for the SUV calculation in seconds.
        end_time: End time for the SUV calculation in seconds.
        
    Returns:
        suv_img (ants.ANTsImage): The standard uptake value calculated on the input image."""
    wss_img = weighted_sum_for_suv(input_image_path=input_image_path,
                                   output_image_path=None,
                                   start_time=start_time,
                                   end_time=end_time)
    suv_img = wss_img / (dose*1000) * weight

    if output_image_path is not None:
        ants.image_write(suv_img, output_image_path)
        safe_copy_meta(input_image_path=input_image_path,
                       out_image_path=output_image_path)

    return suv_img


def suvr(input_image_path: str,
         output_image_path: str | None,
         segmentation_image_path: str,
         ref_region: int | list[int],
         start_time: float,
         end_time: float) -> ants.ANTsImage:
    """
    Computes an ``SUVR`` (Standard Uptake Value Ratio) by taking the average of
    an input image within a reference region, and dividing the input image by
    said average value.

    Args:
        input_image_path (str): Path to 4D PET image.
        output_image_path (str): Path to output image file which is written to. If None, no output
            is written.
        segmentation_image_path (str): Path to segmentation image, which we use
            to compute average uptake value in the reference region.
        ref_region (int): Region or list of region mappings over which to compute average SUV. If a
            list is provided, combines all regions in the list as one reference region.
        start_time: Time in seconds from the start of the scan from which to begin sum calculation.
            Only frames after selected time will be included in the sum. Default 0.
        end_time: Time in seconds from the start of the scan from which to end sum calculation.
            Only frames before selected time will be included in the sum. If -1, use all frames
            after `start_time` in the calculation. Default -1.

    Returns:
        ants.ANTsImage: SUVR parametric image
    """
    sum_img = weighted_sum_for_suv(input_image_path=input_image_path,
                                   output_image_path=None,
                                   start_time=start_time,
                                   end_time=end_time)
    segmentation_img = ants.image_read(filename=segmentation_image_path)

    ref_region_avg = mean_value_in_region(input_img=sum_img,
                                          seg_img=segmentation_img,
                                          mappings=ref_region)

    suvr_img = sum_img / ref_region_avg

    if output_image_path is not None:
        ants.image_write(image=suvr_img,
                         filename=output_image_path)
        safe_copy_meta(input_image_path=input_image_path,
                       out_image_path=output_image_path)

    return suvr_img



"""
Command-line interface (CLI) for interpolating Time-Activity Curves (TACs).

This module provides a CLI to interpolate TACs, enabling the conversion of these curves to a desired time scale. It uses
argparse to handle command-line arguments.

The user must provide:
    * Input TAC file path
    * Interpolation interval in minutes
    * Output TAC file path

This script uses functionalities from :mod:'petpal.tac_interpolation' module to interpolate the TAC based on the
specified interval, or number of samples before the maximum value, and save the interpolated TAC to a specified output
file.

An option to print the interpolated TAC to screen is also provided for visual inspection of the results.

Example usage:
    Using :class:`petpal.cli_tac_interpolation.EvenlyInterpolate`:
        .. code-block:: bash
        
            petpal-tac-interpolation --input-tac-path /path/to/input.tac --output-tac-path /path/to/output.tac --delta-time 0.1
    
    Using :class:`petpal.cli_tac_interpolation.EvenlyInterpolateWithMax`:
        .. code-block:: bash
            
            petpal-tac-interpolation --input-tac-path /path/to/input.tac --output-tac-path /path/to/output.tac --samples-before-max 4
    
See Also:
    :mod:`tac_interpolation <petpal.tac_interpolation>` - module for performing interpolation on PET TACs.

TODO:
    * Refactor the reading and writing of TACs when IO module is mature.
    
"""

import os
import argparse
import numpy as np
from ..kinetic_modeling import tac_interpolation as tac_intp
from ..utils.time_activity_curve import safe_load_tac


def _safe_write_tac(tac_times: np.ndarray, tac_values: np.ndarray, filename: str) -> None:
    """
    Writes time-activity curves (TAC) to a file.

    Tries to write a TAC to the specified file and raises an exception if unable to do so. The TAC is expected to be a
    numpy array where the first index corresponds to the times, and the second corresponds to the activity.

    Args:
        tac_times (np.ndarray): A numpy array containing the time points of the TAC.
        tac_values (np.ndarray): A numpy array containing the activity values of the TAC.
        filename (str): The name of the file to write to.

    Raises:
        Exception: An error occurred writing the TAC.
    """
    out_arr = np.array([tac_times, tac_values]).T
    try:
        np.savetxt(fname=f"{filename}.tsv", X=out_arr, delimiter="\t", header="Time\tActivity", fmt="%.6e")
    except Exception as e:
        print(f"Couldn't write file {filename}. Error: {e}")
        raise e


def _print_tac_to_screen(tac_times: np.ndarray, tac_values: np.ndarray):
    """
    Prints the Time-Activity Curve (TAC) times and values to the console.

    This function takes as input two numpy arrays, one with the TAC times and the other with the TAC values, and prints
    them to the console in a formatted manner. The format is '%.6e\t%.6e'.

    Args:
        tac_times (np.ndarray): A numpy array containing the TAC times.
        tac_values (np.ndarray): A numpy array containing the TAC values.

    """
    print(f"#{'Time':<9}\tActivity")
    for time, value in zip(tac_times, tac_values):
        print(f"{time:<.6e}\t{value:<.6e}")


def main():
    parser = argparse.ArgumentParser(prog="TAC Interpolation", description="Evenly resample TACs.",
                                     epilog="Example of usage: petpal-tac-interpolate -i /path/to/input/file.txt -o "
                                            "/path/to/output/file.txt --delta-time 0.1")

    io_grp = parser.add_argument_group("I/O")
    io_grp.add_argument("-i", "--tac-path", help="Path to TAC file.", required=True)
    io_grp.add_argument("-o", "--out-tac-path", help="Path of output file.", required=True)
    
    interp_grp = parser.add_argument_group("Interpolation")
    mutually_exclusive_group = interp_grp.add_mutually_exclusive_group(required=True)
    mutually_exclusive_group.add_argument("--delta-time", type=float,
                                          help="The time difference for the resampled times.")
    mutually_exclusive_group.add_argument("--samples-before-max", type=float,
                                          help="Number of samples before the max TAC value.")
    
    verb_group = parser.add_argument_group("Additional information")
    verb_group.add_argument("-p", "--print", action="store_true", help="Print the resampled TAC values.",
                            required=False)
    verb_group.add_argument("-v", "--verbose", action="store_true", help="Print the sizes of the input and output TACs",
                            required=False)

    args = parser.parse_args()
    args.tac_path = os.path.abspath(args.tac_path)
    args.out_tac_path = os.path.abspath(args.out_tac_path)

    in_tac_times, in_tac_values = safe_load_tac(args.tac_path)

    if args.samples_before_max is not None:
        interpolator = tac_intp.EvenlyInterpolateWithMax(tac_times=in_tac_times, tac_values=in_tac_values,
                                                         samples_before_max=args.samples_before_max)
    else:
        interpolator = tac_intp.EvenlyInterpolate(tac_times=in_tac_times, tac_values=in_tac_values,
                                                  delta_time=args.delta_time)

    resampled_times, resampled_values = interpolator.get_resampled_tac()

    _safe_write_tac(tac_times=resampled_times, tac_values=resampled_values, filename=args.out_tac_path)

    if args.verbose:
        print(f"Input TAC size:  {len(in_tac_values)}.")
        print(f"Output TAC size: {len(resampled_values)}.")

    if args.print:
        _print_tac_to_screen(tac_times=resampled_times, tac_values=resampled_values)


if __name__ == "__main__":
    main()



import pathlib
import os
import copy
from typing import Union
from .steps_base import *
from .steps_containers import StepsContainer, StepsPipeline
from ..utils.bids_utils import gen_bids_like_dir_path, gen_bids_like_filepath


class BIDSyPathsForRawData:
    """
    A class to manage and generate paths for raw BIDS data and its derivatives.

    This class handles the generation and validation of paths for different types of raw data
    (such as PET, anatomical images, and blood TAC files) and their derivatives, following the
    BIDS format.

    Attributes:
        sub_id (str): Subject ID.
        ses_id (str): Session ID.
        _bids_dir (Optional[str]): Root directory for BIDS data.
        _derivatives_dir (Optional[str]): Directory for derivative data.
        _raw_pet_path (Optional[str]): Path for raw PET images.
        _raw_anat_path (Optional[str]): Path for raw anatomical images.
        _segmentation_img_path (Optional[str]): Path for segmentation images.
        _segmentation_label_table_path (Optional[str]): Path for segmentation label tables.
        _raw_blood_tac_path (Optional[str]): Path for raw blood TAC files.
    """
    def __init__(self,
                 sub_id: str,
                 ses_id: str,
                 bids_root_dir: str = None,
                 derivatives_dir: str = None,
                 raw_pet_img_path: str = None,
                 raw_anat_img_path: str = None,
                 segmentation_img_path: str = None,
                 segmentation_label_table_path: str = None,
                 raw_blood_tac_path: str = None,):
        """
        Initializes the BIDSyPathsForRawData object with the given subject and session IDs,
        and paths for various types of data.

        Args:
            sub_id (str): Subject ID.
            ses_id (str): Session ID.
            bids_root_dir (Optional[str]): Optional path to the BIDS root directory. Defaults to None.
            derivatives_dir (Optional[str]): Optional path to the derivatives directory. Defaults to None.
            raw_pet_img_path (Optional[str]): Optional path to the raw PET image. Defaults to None.
            raw_anat_img_path (Optional[str]): Optional path to the raw anatomical image. Defaults to None.
            segmentation_img_path (Optional[str]): Optional path to the segmentation image. Defaults to None.
            segmentation_label_table_path (Optional[str]): Optional path to the segmentation label table. Defaults to None.
            raw_blood_tac_path (Optional[str]): Optional path to the raw blood TAC file. Defaults to None.
        """
        self.sub_id = sub_id
        self.ses_id = ses_id
        self._bids_dir = bids_root_dir
        self._derivatives_dir = derivatives_dir
        self._raw_pet_path = raw_pet_img_path
        self._raw_anat_path = raw_anat_img_path
        self._segmentation_img_path = segmentation_img_path
        self._segmentation_label_table_path = segmentation_label_table_path
        self._raw_blood_tac_path = raw_blood_tac_path
        
        self.bids_dir = self._bids_dir
        self.derivatives_dir = self._derivatives_dir
        self.pet_path = self._raw_pet_path
        self.anat_path = self._raw_anat_path
        self.seg_img = self._segmentation_img_path
        self.seg_table = self._segmentation_label_table_path
        self.blood_path = self._raw_blood_tac_path
        
    def __repr__(self):
        """
        Provides an unambiguous string representation of the BIDSyPathsForRawData instance.

        Returns:
            str: A string representation showing how the instance can be recreated.
        """
        type_name = type(self).__name__
        name_len = len(type_name)+1
        info_str = [f'{type(self).__name__}(sub_id={repr(self.sub_id)}',
                    f'ses_id={repr(self.ses_id)}',
                    f'bids_root_dir={repr(self.bids_dir)}',
                    f'derivatives_dir={repr(self.derivatives_dir)}',
                    f'raw_pet_img_path={repr(self.pet_path)}',
                    f'raw_anat_img_path={repr(self.anat_path)}',
                    f'segmentation_img_path={repr(self.seg_img)}',
                    f'segmentation_label_table_path={repr(self.seg_table)}',
                    f'raw_blood_tac_path={repr(self.blood_path)}',
                    f')']
        return (",\n"+" "*name_len).join(info_str)
    
    def __str__(self):
        """
        Returns a human-readable string representation of the object.

        Returns:
            str: A string representation of the object.
        """
        info_str = ["(BIDSyPaths info):",
                    f"Subject ID: {self.sub_id}",
                    f"Session ID: {self.ses_id}",
                    f"BIDS root dir: {self.bids_dir}",
                    f"Derivatives dir: {self.derivatives_dir}",
                    f"4D-PET image: {repr(self.pet_path)}",
                    f"Anatomical image: {repr(self.anat_path)}",
                    f"Segmentation image: {repr(self.seg_img)}",
                    f"Segmentation table: {repr(self.seg_table)}",
                    f"Blood TAC: {repr(self.blood_path)}",]
        return "\n".join(info_str)
    
    
    @property
    def bids_dir(self) -> str:
        """
        str: Property for the BIDS root directory.
        """
        return self._bids_dir
    
    @bids_dir.setter
    def bids_dir(self, value):
        """
        Sets the BIDS root directory, validating if it is a directory.

        Args:
            value (Optional[str]): Path to the BIDS root directory.

        Raises:
            ValueError: If the given path is not a directory.
        """
        if value is None:
            self._bids_dir = os.path.abspath('../')
        else:
            val_path = pathlib.Path(value)
            if val_path.is_dir():
                self._bids_dir = os.path.abspath(value)
            else:
                raise ValueError("Given BIDS path is not a directory.")
            
    @property
    def derivatives_dir(self):
        """
        str: Property for the derivatives directory.
        """
        return self._derivatives_dir
    
    @derivatives_dir.setter
    def derivatives_dir(self, value):
        """
        Sets the derivatives directory, validating if it is a sub-directory of the BIDS root directory.

        Args:
            value (Optional[str]): Path to the derivatives directory.

        Raises:
            ValueError: If the given path is not a sub-directory of the BIDS root directory.
        """
        if value is None:
            self._derivatives_dir = os.path.abspath(os.path.join(self.bids_dir, 'derivatives'))
        else:
            val_path = pathlib.Path(value).absolute()
            if val_path.is_relative_to(self.bids_dir):
                self._derivatives_dir = os.path.abspath(value)
            else:
                raise ValueError(f"Given derivatives path is not a sub-directory of BIDS path."
                                 f"\nBIDS:       {self.bids_dir}"
                                 f"\nDerivatives:{os.path.abspath(value)}")
    
    
    @property
    def pet_path(self):
        """
        str: Property for the raw PET image path.
        """
        return self._raw_pet_path
    
    @pet_path.setter
    def pet_path(self, value: str):
        """
        Sets the raw PET image path, generating a path if None provided, and validating the file.

        Args:
            value (Optional[str]): Path to the raw PET image. If None, a path will be generated.

        Raises:
            FileNotFoundError: If the given file does not exist or does not have a .nii.gz extension.
        """
        if value is None:
            filepath = gen_bids_like_filepath(sub_id=self.sub_id,
                                              ses_id=self.ses_id,
                                              modality='pet',
                                              bids_dir=self.bids_dir,
                                              ext='.nii.gz')
            self._raw_pet_path = filepath
        else:
            val_path = pathlib.Path(value)
            val_suff = "".join(val_path.suffixes)
            if val_path.is_file() and val_suff in ('.nii.gz','.nii'):
                self._raw_pet_path = value
            else:
                raise FileNotFoundError(f"File does not exist: {value}")
            
    @property
    def anat_path(self):
        """
        str: Property for the raw anatomical image path.
        """
        return self._raw_anat_path
    
    @anat_path.setter
    def anat_path(self, value: str):
        """
        Sets the raw anatomical image path, generating a path if None provided, and validating the file.

        Args:
            value (Optional[str]): Path to the raw anatomical image. If None, a path will be generated.

        Raises:
            FileNotFoundError: If the given file does not exist or does not have a .nii.gz extension.
        """
        if value is None:
            filepath = gen_bids_like_filepath(sub_id=self.sub_id, ses_id=self.ses_id,
                                              modality='anat', suffix='MPRAGE',
                                              bids_dir=self.bids_dir, ext='.nii.gz')
            self._raw_anat_path =  filepath
        else:
            val_path = pathlib.Path(value)
            val_suff = "".join(val_path.suffixes)
            if val_path.is_file() and val_suff in ('.nii.gz','.nii'):
                self._raw_anat_path = value
            else:
                raise FileNotFoundError(f"File does not exist: {value}")
            
    @property
    def seg_img(self):
        """
        str: Property for the segmentation image path.
        """
        return self._segmentation_img_path
    
    @seg_img.setter
    def seg_img(self, value: str):
        """
        Sets the segmentation image path, generating a path if None provided, and validating the file.

        Args:
            value (Optional[str]): Path to the segmentation image. If None, a path will be generated.

        Raises:
            FileNotFoundError: If the given file does not exist.
        """
        if value is None:
            seg_dir = os.path.join(self.derivatives_dir, 'ROI_mask')
            filepath = gen_bids_like_filepath(sub_id=self.sub_id, ses_id=self.ses_id,
                                              modality='anat', bids_dir=seg_dir,
                                              suffix='ROImask',
                                              ext='.nii.gz',
                                              space='MPRAGE',
                                              desc='lesionsincluded')
            self._segmentation_img_path = filepath
        else:
            if os.path.isfile(value):
                self._segmentation_img_path = value
            else:
                raise FileNotFoundError(f"File does not exist: {value}")
            
    @property
    def seg_table(self):
        """
        str: Property for the segmentation label table path.
        """
        return self._segmentation_label_table_path
    
    @seg_table.setter
    def seg_table(self, value: str):
        """
        Sets the segmentation label table path, generating a path if None provided, and validating the file.

        Args:
            value (Optional[str]): Path to the segmentation label table. If None, a path will be generated.

        Raises:
            FileNotFoundError: If the given file does not exist or does not have a .tsv extension.
        """
        if value is None:
            seg_dir = os.path.join(self.derivatives_dir, 'ROI_mask')
            filename = 'dseg.tsv'
            self._segmentation_label_table_path = os.path.join(seg_dir, filename)
        else:
            val_path = pathlib.Path(value)
            if val_path.is_file() and (val_path.suffix == '.tsv'):
                self._segmentation_label_table_path = value
            else:
                raise FileNotFoundError(f"File does not exist: {value}")
            
    @property
    def blood_path(self):
        """
        str: Property for the raw blood TAC path.
        """
        return self._raw_blood_tac_path
    
    @blood_path.setter
    def blood_path(self, value: str):
        """
        Sets the raw blood TAC path, generating a path if None provided, and validating the file.

        Args:
            value (Optional[str]): Path to the raw blood TAC file. If None, a path will be generated.

        Raises:
            FileNotFoundError: If the given file does not exist or does not have a .tsv extension.
        """
        if value is None:
            filepath = gen_bids_like_filepath(sub_id=self.sub_id,
                                              ses_id=self.ses_id,
                                              bids_dir=self.bids_dir,
                                              modality='pet',
                                              suffix='blood',
                                              ext='.tsv',
                                              desc='decaycorrected'
                                              )
            self._raw_blood_tac_path = filepath
        else:
            val_path = pathlib.Path(value)
            if val_path.is_file() and (val_path.suffix == '.tsv'):
                self._raw_blood_tac = value
            else:
                raise FileNotFoundError(f"File does not exist: {value}")
                
                
class BIDSyPathsForPipelines(BIDSyPathsForRawData):
    """
    A class to manage and generate paths for BIDS data and pipeline-specific analysis directories.

    Inherits from :class:`BIDSyPathsForRawData` and adds functionality to handle pipeline-specific directories
    and analysis results.

    Attributes:
        sub_id (str): Subject ID.
        ses_id (str): Session ID.
        pipeline_name (str): Name of the pipeline. Defaults to 'generic_pipeline'.
        list_of_analysis_dir_names (Union[None, list[str]]): List of names for analysis directories.
        _pipeline_dir (Optional[str]): Directory for the pipeline.
        analysis_dirs (dict): A dictionary containing paths to analysis directories.
    """
    def __init__(self,
                 sub_id: str,
                 ses_id: str,
                 pipeline_name: str ='generic_pipeline',
                 list_of_analysis_dir_names: Union[None, list[str]] = None,
                 bids_root_dir: str = None,
                 derivatives_dir: str = None,
                 raw_pet_img_path: str = None,
                 raw_anat_img_path: str = None,
                 segmentation_img_path: str = None,
                 segmentation_label_table_path: str = None,
                 raw_blood_tac_path: str = None):
        """
        Initializes the BIDSyPathsForPipelines object with the given subject and session IDs,
        pipeline name, and paths for various types of data.

        Args:
            sub_id (str): Subject ID.
            ses_id (str): Session ID.
            pipeline_name (str, optional): Name of the pipeline. Defaults to 'generic_pipeline'.
            list_of_analysis_dir_names (Union[None, list[str]], optional): List of names for analysis directories. Defaults to None.
            bids_root_dir (Optional[str]): Optional path to the BIDS root directory. Defaults to None.
            derivatives_dir (Optional[str]): Optional path to the derivatives directory. Defaults to None.
            raw_pet_img_path (Optional[str]): Optional path to the raw PET image. Defaults to None.
            raw_anat_img_path (Optional[str]): Optional path to the raw anatomical image. Defaults to None.
            segmentation_img_path (Optional[str]): Optional path to the segmentation image. Defaults to None.
            segmentation_label_table_path (Optional[str]): Optional path to the segmentation label table. Defaults to None.
            raw_blood_tac_path (Optional[str]): Optional path to the raw blood TAC file. Defaults to None.
        """
        super().__init__(sub_id=sub_id,
                         ses_id=ses_id,
                         bids_root_dir=bids_root_dir,
                         derivatives_dir=derivatives_dir,
                         raw_pet_img_path=raw_pet_img_path,
                         raw_anat_img_path=raw_anat_img_path,
                         segmentation_img_path=segmentation_img_path,
                         segmentation_label_table_path=segmentation_label_table_path,
                         raw_blood_tac_path=raw_blood_tac_path)
        
        self._pipeline_dir = None
        self.pipeline_name = pipeline_name
        self.pipeline_dir = self._pipeline_dir
        self.list_of_analysis_dir_names = list_of_analysis_dir_names
        self.analysis_dirs = self.generate_analysis_dirs(list_of_dir_names=list_of_analysis_dir_names)
        self.make_analysis_dirs()
        
    def __repr__(self):
        """
        Provides an unambiguous string representation of the BIDSyPathsForPipelines instance.

        Returns:
            str: A string representation showing how the instance can be recreated.
            
        See Also:
            :func:`BIDSyPathsForRawData.__repr__`
        """
        type_name = type(self).__name__
        name_len = len(type_name) + 1
        info_str = super().__repr__().split("\n")
        info_str[0] = f'{type(self).__name__}(sub_id={repr(self.sub_id)},'
        info_str.insert(2," "*name_len+f"pipeline_name={repr(self.pipeline_name)},")
        info_str.insert(3," "*name_len+f"list_of_analysis_dir_names={repr(self.list_of_analysis_dir_names)},")
        return ("\n").join(info_str)
    
    def __str__(self):
        """
        Returns a human-readable string representation of the object.

        Returns:
            str: A string representation of the object.
        """
        info_str = super().__str__().split("\n")
        info_str.append(f"Pipeline Name: {self.pipeline_name}")
        info_str.append(f"Pipeline Dir: {self.pipeline_dir}")
        info_str.append(f"Analysis Dirs:")
        for dir_name, dir_path in self.analysis_dirs.items():
            info_str.append(f"\t{dir_name}: {dir_path}")
            
        return "\n".join(info_str)
        
        
    @property
    def pipeline_dir(self):
        """
        str: Property for the pipeline directory.
        """
        return self._pipeline_dir
    
    @pipeline_dir.setter
    def pipeline_dir(self, value: str):
        """
        Sets the pipeline directory, validating if it is a sub-directory of the derivatives directory.

        Args:
            value (Optional[str]): Path to the pipeline directory. If None, a default path will be generated.

        Raises:
            ValueError: If the pipeline directory is not relative to the derivatives directory.
        """
        if value is None:
            default_path = os.path.join(self.derivatives_dir, 'petpal', 'pipelines', self.pipeline_name)
            self._pipeline_dir = os.path.abspath(default_path)
        else:
            pipe_path = pathlib.Path(value).absolute()
            if pipe_path.is_relative_to(self.derivatives_dir):
                self._pipeline_dir = os.path.abspath(value)
            else:
                raise ValueError("Pipeline directory is not relative to the derivatives directory")
    
    def generate_analysis_dirs(self, list_of_dir_names: Union[None, list[str]] = None) -> dict:
        """
        Generates paths for analysis directories.

        Args:
            list_of_dir_names (Union[None, list[str]], optional): List of names for analysis directories.
                Defaults to None, in which case default directories will be used.

        Returns:
            dict: A dictionary containing paths to analysis directories.
        """
        if list_of_dir_names is None:
            list_of_dir_names = ['preproc', 'km', 'tacs']
        path_gen = lambda name: gen_bids_like_dir_path(sub_id=self.sub_id,
                                                       ses_id=self.ses_id,
                                                       modality=name,
                                                       sup_dir=self.pipeline_dir)
        analysis_dirs = {name:path_gen(name) for name in list_of_dir_names}
        return analysis_dirs
    
    def make_analysis_dirs(self):
        """
        Creates the analysis directories if they do not already exist.
        """
        for a_name, a_dir in self.analysis_dirs.items():
            os.makedirs(a_dir, exist_ok=True)
            
    
class BIDS_Pipeline(BIDSyPathsForPipelines, StepsPipeline):
    """
    A class that combines BIDS data path management with a steps-based pipeline for processing.

    Inherits from:
        BIDSyPathsForPipelines: Manages paths for BIDS data and analysis directories.
        StepsPipeline: Manages a series of processing steps in a pipeline.

    Attributes:
        sub_id (str): Subject ID.
        ses_id (str): Session ID.
        pipeline_name (str): Name of the pipeline. Defaults to 'generic_pipeline'.
        list_of_analysis_dir_names (Union[None, list[str]]): List of names for analysis directories.
        bids_root_dir (Optional[str]): Optional path to the BIDS root directory.
        derivatives_dir (Optional[str]): Optional path to the derivatives directory.
        raw_pet_img_path (Optional[str]): Optional path to the raw PET image.
        raw_anat_img_path (Optional[str]): Optional path to the raw anatomical image.
        segmentation_img_path (Optional[str]): Optional path to the segmentation image.
        segmentation_label_table_path (Optional[str]): Optional path to the segmentation label table.
        raw_blood_tac_path (Optional[str]): Optional path to the raw blood TAC file.
        step_containers (list[StepsContainer]): List of step containers for the pipeline.
        
    Example:
        
        The following is a basic example demonstrating how to instantiate the basic default
        pipeline which performs the following steps:

        #. Crop the raw PET image using a threshold.
        #. Motion correct the cropped image to a mean-PET target where the frames have mean intensities
           greater than or equal to the total mean intensity.
        #. Computes a `weighted-series-sum` image from the cropped PET image. **We add this step since
           it is not part of the default pipeline.**
        #. Registers the motion corrected PET image to a T1w reference image.
        #. For each of the ROI segments defined, we extract TACs and save them.
        #. For the blood TAC, which is assumed to be decay corrected (and WB corrected if appliclable),
           we resample the TAC on the PET scan frame times.
        #. Generate parametric patlak slope and intercept images from the register PET image.
        #. Generate parametric logan slope and intercept images from the register PET image.
        #. For each ROI TAC, calculate a 1TCM fit.
        #. For each ROI TAC, calculate a 1TCM fit.
        #. For each ROI TAC, calculate an irreversible 2TCM (:math:`k_{4}=0`) fit.
        #. For each ROI TAC, calculate a serial 2TCM fit.
        #. For each ROI TAC, calculate a patlak fit.
        #. For each ROI TAC, calculate a logan fit.
        
        We assume that we are running the following code in the ``/code`` folder of a BIDS project.
        
        .. code-block:: python
        
            from petpal.pipelines.pipelines import *
            
            
            # Assuming that the current directory is the `BIDS_ROOT/code` of a BIDS directory.
            
            this_pipeline = BIDS_Pipeline.default_bids_pipeline(sub_id='XXXX',
                                                                ses_id='XX',
                                                                pipeline_name='study_pipeline')
            
            # Plot the dependency graph to quickly glance at all the steps in the pipeline.
            this_pipeline.plot_dependency_graph()
            
            # Check if all the steps can potentially run
            print(this_pipeline.can_steps_potentially_run())
            
            # Check which steps can potentially run
            print(this_pipeline.get_steps_potential_run_state())
            
            ## Editing a pipeline by adding new steps and removing pre-defined steps
            # Instantiating a weighted-series-sum step, and using the pipeline-inferred raw-PET image
            # path to infer the half-life of the radioisotope for the calculation.
            from petpal.pipelines.preproc_steps import ImageToImageStep
            from petpal.utils.useful_functions import weighted_series_sum as wss_func
            wss_step = ImageToImageStep(name='wss',
                                        function=wss_func,
                                        input_image_path='',
                                        output_image_path='',
                                        half_life=get_half_life_from_nifti(this_pipeline.pet_path),
                                        verbose=False
                                       )
            
            # Adding the step to the pipeline with the dependency that wss receives the output from
            # the 'thresh_crop' step in the 'preproc' container.
            this_pipeline.add_step(container_name='preproc', step=wss_step)
            this_pipeline.add_dependency(sending='thresh_crop', receiving='wss')
            
            # Removing the step for calculating the Alt-Logan fits for each of the ROI TACs
            this_pipeline.remove_step('roi_alt_logan_fit')
            
            # Removing the step for calculating the parametric Alt-Logan fits.
            this_pipeline.remove_step('parametric_alt_logan_fit')
            
            # Since we added, and removed steps, we have to update the dependencies.
            this_pipeline.update_dependencies()
            
            # Looking at the updated dependency graph in text-format
            this_pipeline.print_dependency_graph()
            
            # Run all the steps in the pipeline in topological order.
            this_pipeline()
        
    """
    def __init__(self,
                 sub_id: str,
                 ses_id: str,
                 pipeline_name: str = 'generic_pipeline',
                 list_of_analysis_dir_names: Union[None, list[str]] = None,
                 bids_root_dir: str = None,
                 derivatives_dir: str = None,
                 raw_pet_img_path: str = None,
                 raw_anat_img_path: str = None,
                 segmentation_img_path: str = None,
                 segmentation_label_table_path: str = None,
                 raw_blood_tac_path: str = None,
                 step_containers: list[StepsContainer] = []):
        """
        Initializes the BIDS_Pipeline object with the given subject and session IDs,
        pipeline name, paths for various types of data, and step containers for processing.

        Args:
            sub_id (str): Subject ID.
            ses_id (str): Session ID.
            pipeline_name (str, optional): Name of the pipeline. Defaults to 'generic_pipeline'.
            list_of_analysis_dir_names (Union[None, list[str]], optional): List of names for analysis
                directories. Defaults to None.
            bids_root_dir (Optional[str]): Optional path to the BIDS root directory. Defaults to None.
            derivatives_dir (Optional[str]): Optional path to the derivatives directory. Defaults to None.
            raw_pet_img_path (Optional[str]): Optional path to the raw PET image. Defaults to None.
            raw_anat_img_path (Optional[str]): Optional path to the raw anatomical image. Defaults to None.
            segmentation_img_path (Optional[str]): Optional path to the segmentation image. Defaults to None.
            segmentation_label_table_path (Optional[str]): Optional path to the segmentation label table.
                Defaults to None.
            raw_blood_tac_path (Optional[str]): Optional path to the raw blood TAC file. Defaults to None.
            step_containers (list[StepsContainer], optional): List of step containers for the pipeline.
                Defaults to an empty list.
        """
        BIDSyPathsForPipelines.__init__(self,
                                        sub_id=sub_id,
                                        ses_id=ses_id,
                                        pipeline_name=pipeline_name,
                                        list_of_analysis_dir_names=list_of_analysis_dir_names,
                                        bids_root_dir=bids_root_dir,
                                        derivatives_dir=derivatives_dir,
                                        raw_pet_img_path=raw_pet_img_path,
                                        raw_anat_img_path=raw_anat_img_path,
                                        segmentation_img_path=segmentation_img_path,
                                        segmentation_label_table_path=segmentation_label_table_path,
                                        raw_blood_tac_path=raw_blood_tac_path)
        StepsPipeline.__init__(self, name=pipeline_name, step_containers=step_containers)
        

    def __repr__(self):
        """
        Provides an unambiguous string representation of the BIDSyPathsForPipelines instance.

        Returns:
            str: A string representation showing how the instance can be recreated.

        See Also:
            - :meth:`BIDSyPathsForPipelines.__repr__`
            - :meth:`StepsPipeline.__repr__`
            
        """
        cls_name = type(self).__name__
        info_str = [f'{cls_name}(', ]
        
        in_kwargs = ArgsDict(dict(sub_id=self.sub_id,
                                  ses_id=self.ses_id,
                                  pipeline_name = self.name,
                                  list_of_analysis_dir_names = self.list_of_analysis_dir_names,
                                  bids_root_dir = self.bids_dir,
                                  derivatives_dir = self.derivatives_dir,
                                  raw_pet_img_path = self.pet_path,
                                  raw_anat_img_path = self.anat_path,
                                  segmentation_img_path = self.seg_img,
                                  segmentation_label_table_path = self.seg_table,
                                  raw_blood_tac_path = self.blood_path)
                )
        
        for arg_name, arg_val in in_kwargs.items():
            info_str.append(f'{arg_name}={repr(arg_val)},')
        
        info_str.append('step_containers=[')
        
        for _, container in self.step_containers.items():
            info_str.append(f'{repr(container)},')
        
        info_str.append(']')
        info_str.append(')')
        
        return f'\n    '.join(info_str)
    
    def __str__(self):
        """
        Returns a human-readable string representation of the object. Lists all the directories
        and paths, and the steps-containers with their dependencies.

        Returns:
            str: A string representation of the object.
            
        See Also:
            - :meth:`BIDSyPathsForPipelines.__str__`
            - :meth:`StepsPipeline.__str__<petpal.pipelines.steps_contaiers.StepsPipeline.__str__>`
        """
        pipeline_info_str = StepsPipeline.__str__(self).split("\n")
        paths_info_str = BIDSyPathsForPipelines.__str__(self).split("\n")
        info_str = ["*"*50]+paths_info_str + pipeline_info_str
        return "\n".join(info_str)
        
        
    def update_dependencies_for(self, step_name, verbose=False):
        """
        Updates the dependencies for a specified step in the pipeline. Extends
        :meth:`StepsPipeline.update_dependencies_for<petpal.pipelines.steps_containers.StepsPipeline.update_dependencies_for>`
        to infer the outputs from input paths after updating dependencies.

        Args:
            step_name (str): The name of the step for which to update dependencies.
            verbose (bool, optional): If True, print verbose updates. Defaults to False.
            
        Raises:
            NotImplementedError: The ``infer_outputs_from_inputs`` for the sending step is not
                implemented, OR, ``set_input_as_output_from`` for the receiving step is not
                implemented.
        
        """
        super().update_dependencies_for(step_name, verbose=verbose)
        this_step = self.get_step_from_node_label(step_name)
        this_step_grp_name = self.dependency_graph.nodes(data=True)[step_name]['grp']
        try:
            this_step.infer_outputs_from_inputs(out_dir=self.pipeline_dir,
                                                der_type=this_step_grp_name,)
        except NotImplementedError:
            raise NotImplementedError(f"Step {step_name} does not have the `infer_outputs_from_inputs` "
                                      f"method implemented.")

    @classmethod
    def default_bids_pipeline(cls,
                              sub_id: str,
                              ses_id: str,
                              pipeline_name: str = 'generic_pipeline',
                              list_of_analysis_dir_names: Union[None, list[str]] = None,
                              bids_root_dir: str = None,
                              derivatives_dir: str = None,
                              raw_pet_img_path: str = None,
                              raw_anat_img_path: str = None,
                              segmentation_img_path: str = None,
                              segmentation_label_table_path: str = None,
                              raw_blood_tac_path: str = None):
        """
        Creates a default BIDS pipeline with predefined steps and dependencies.

        Args:
            sub_id (str): Subject ID.
            ses_id (str): Session ID.
            pipeline_name (str, optional): Name of the pipeline. Defaults to 'generic_pipeline'.
            list_of_analysis_dir_names (Union[None, list[str]], optional): List of names for
                analysis directories. Defaults to None.
            bids_root_dir (Optional[str]): Optional path to the BIDS root directory. Defaults
                to None.
            derivatives_dir (Optional[str]): Optional path to the derivatives directory.
                Defaults to None.
            raw_pet_img_path (Optional[str]): Optional path to the raw PET image. Defaults
                to None.
            raw_anat_img_path (Optional[str]): Optional path to the raw anatomical image.
                Defaults to None.
            segmentation_img_path (Optional[str]): Optional path to the segmentation image.
                Defaults to None.
            segmentation_label_table_path (Optional[str]): Optional path to the segmentation
                label table. Defaults to None.
            raw_blood_tac_path (Optional[str]): Optional path to the raw blood TAC file.
                Defaults to None.

        Returns:
            BIDS_Pipeline: A BIDS_Pipeline object with the default steps and dependencies set.
            
        Notes:
            The following steps are defined:
                - Crop the raw PET image using a threshold.
                - Motion correct the cropped image to a mean-PET target where the frames have mean intensities
                  greater than or equal to the total mean intensity.
                - Registers the motion corrected PET image to a T1w reference image.
                - For each of the ROI segments defined, we extract TACs and save them.
                - For the blood TAC, which is assumed to be decay corrected (and WB corrected if appliclable),
                  we resample the TAC on the PET scan frame times.
                - Generate parametric patlak slope and intercept images from the register PET image.
                - Generate parametric logan slope and intercept images from the register PET image.
                - Generate parametric alt-logan slope and intercept images from the register PET image.
                - For each ROI TAC, calculate a 1TCM fit.
                - For each ROI TAC, calculate a 1TCM fit.
                - For each ROI TAC, calculate an irreversible 2TCM (:math:`k_{4}=0`) fit.
                - For each ROI TAC, calculate a serial 2TCM fit.
                - For each ROI TAC, calculate a patlak fit.
                - For each ROI TAC, calculate a logan fit.
                - For each ROI TAC, calculate an alt-logan fit.
                
        See Also:
            - :meth:`default_preprocess_steps<petpal.pipelines.steps_containers.StepsContainer.default_preprocess_steps>`
            - :meth:`default_kinetic_analysis_steps<petpal.pipelines.steps_containers.StepsContainer.default_kinetic_analysis_steps>`
            
        """
        temp_pipeline = StepsPipeline.default_steps_pipeline()
        
        obj = cls(sub_id=sub_id,
                  ses_id=ses_id,
                  pipeline_name=pipeline_name,
                  list_of_analysis_dir_names=list_of_analysis_dir_names,
                  bids_root_dir=bids_root_dir,
                  derivatives_dir=derivatives_dir,
                  raw_pet_img_path=raw_pet_img_path,
                  raw_anat_img_path=raw_anat_img_path,
                  segmentation_img_path=segmentation_img_path,
                  segmentation_label_table_path=segmentation_label_table_path,
                  raw_blood_tac_path=raw_blood_tac_path,
                  step_containers=list(temp_pipeline.step_containers.values())
                  )
        
        obj.dependency_graph = copy.deepcopy(temp_pipeline.dependency_graph)
        
        del temp_pipeline
        
        containers = obj.step_containers
        
        containers["preproc"][0].input_image_path = obj.pet_path
        containers["preproc"][2].kwargs['reference_image_path'] = obj.anat_path
        containers["preproc"][3].segmentation_label_map_path = obj.seg_table
        containers["preproc"][3].segmentation_image_path = obj.seg_img
        containers["preproc"][4].raw_blood_tac_path = obj.blood_path
        
        obj.update_dependencies(verbose=False)
        return obj



import inspect
from typing import Callable


class ArgsDict(dict):
    """
    A specialized subclass of Python's built-in `dict` that provides a customized string representation.

    Attributes:
        None, since ArgsDict inherits directly from dict.
    """
    def __str__(self):
        """
        Returns a formatted string representation of the dictionary.
        
        The string output will list each key-value pair on a new line with indentation to improve readability.
        
        Returns:
            str: A string containing the formatted key-value pairs, indented for clarity.
        """
        rep_str = [f'    {arg}={val}' for arg, val in self.items()]
        return ',\n'.join(rep_str)


class StepsAPI:
    """
    StepsAPI provides an interface for defining steps in a processing pipeline.

    This class outlines methods that allow input and output management between different steps,
    and perform inference of output files based on input data and given parameters.

    """

    def __init__(self, skip_step: bool = False):
        self.skip_step = skip_step

    def execute(self, *args, **kwargs):
        raise NotImplementedError

    def set_input_as_output_from(self, *sending_steps):
        """
        Sets the input of the current step as the output from a list of steps.

        Args:
            *sending_steps: The previous steps from which the output will be used as input for the current step.

        Raises:
            NotImplementedError: This method should be implemented by subclasses.
            
        Notes:
            For a concrete example, take a look at:
            :meth:`TACsFromSegmentationStep<petpal.pipelines.preproc_steps.TACsFromSegmentationStep.set_input_as_output_from>`

        .. important::
           If a step takes multiple input steps. the implementation will have a defined order for steps.

        """
        raise NotImplementedError
    
    def infer_outputs_from_inputs(self, out_dir: str, der_type: str, suffix: str = None, ext: str = None, **extra_desc):
        """
        Infers output files from input data based on the specified output directory,
        derivative type, optional suffix and extension, plus any extra descriptions.

        Args:
            out_dir (str): The directory where the output files will be saved.
            der_type (str): The type of derivative being produced.
            suffix (str, optional): An optional suffix for the output files.
            ext (str, optional): An optional extension for the output files.
            **extra_desc: Additional keyword arguments for extra descriptions to be included.

        Raises:
            NotImplementedError: This method should be implemented by subclasses.
            
        Notes:
            For a concrete example, take a look at:
            :meth:`TACsFromSegmentationStep<petpal.pipelines.preproc_steps.TACsFromSegmentationStep.infer_outputs_from_inputs>`
        """
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        if not self.skip_step:
            self.execute(*args, **kwargs)



class FunctionBasedStep(StepsAPI):
    """
    A step in a processing pipeline based on a callable function.

    This class allows for the execution of a given function with specified arguments and keyword arguments,
    validating that all mandatory parameters are provided.

    Attributes:
        name (str): The name of the step.
        function (Callable): The function to be executed in this step.
        args (tuple): Positional arguments to be passed to the function.
        kwargs (ArgsDict): Keyword arguments to be passed to the function.
        func_sig (inspect.Signature): The signature of the function for validating arguments.

    """
    def __init__(self, name: str, function: Callable, *args, **kwargs) -> None:
        """
        Initializes a function-based step in the processing pipeline.

        Args:
            name (str): The name of the step.
            function (Callable): The function to be executed in this step.
            *args: Positional arguments to be passed to the function.
            **kwargs: Keyword arguments to be passed to the function.
            
        """
        StepsAPI.__init__(self, skip_step=False)
        self.name = name
        self.function = function
        self._func_name = function.__name__
        self.args = args
        self.kwargs = ArgsDict(kwargs)
        self.func_sig = inspect.signature(self.function)
        self.validate_kwargs_for_non_default_have_been_set()
    
    def get_function_args_not_set_in_kwargs(self) -> ArgsDict:
        """
        Retrieves arguments of the function that are not set in the keyword arguments.

        Returns:
            ArgsDict: A dictionary of function arguments that have not been set in the keyword arguments.
        """
        unset_args_dict = ArgsDict()
        func_params = self.func_sig.parameters
        arg_names = list(func_params)
        for arg_name in arg_names[len(self.args):]:
            if arg_name not in self.kwargs:
                if (func_params[arg_name].kind == inspect.Parameter.POSITIONAL_OR_KEYWORD):
                    unset_args_dict[arg_name] = func_params[arg_name].default
        return unset_args_dict
    
    def get_empty_default_kwargs(self) -> list:
        """
        Identifies arguments that have not been provided and lack default values.

        Returns:
            list: A list of argument names that have no default values and are not provided.
        """
        unset_args_dict = self.get_function_args_not_set_in_kwargs()
        empty_kwargs = []
        for arg_name, arg_val in unset_args_dict.items():
            if arg_val is inspect.Parameter.empty:
                if arg_name not in self.kwargs:
                    empty_kwargs.append(arg_name)
        return empty_kwargs
    
    def validate_kwargs_for_non_default_have_been_set(self) -> None:
        """
        Validates that all mandatory arguments have been provided.

        Raises:
            RuntimeError: If any mandatory arguments are missing.
        """
        empty_kwargs = self.get_empty_default_kwargs()
        if empty_kwargs:
            unset_args = '\n'.join(empty_kwargs)
            raise RuntimeError(f"For {self._func_name}, the following arguments must be set:\n{unset_args}")
    
    def execute(self):
        """
        Executes the function with the provided arguments and keyword arguments.

        Raises:
            The function may raise any exceptions that its implementation can throw.
        """
        print(f"(Info): Executing {self.name}")
        self.function(*self.args, **self.kwargs)
        print(f"(Info): Finished {self.name}")
    
    def generate_kwargs_from_args(self) -> ArgsDict:
        """
        Converts positional arguments into keyword arguments.

        Returns:
            ArgsDict: A dictionary where positional arguments are mapped to their corresponding parameter names.
        """
        args_to_kwargs_dict = ArgsDict()
        for arg_name, arg_val in zip(list(self.func_sig.parameters), self.args):
            args_to_kwargs_dict[arg_name] = arg_val
        return args_to_kwargs_dict
    
    def __str__(self):
        """
        Returns a detailed string representation of the FunctionBasedStep instance.

        Returns:
            str: A string describing the step, including its name, function, arguments, and keyword arguments.
        """
        args_to_kwargs_dict = self.generate_kwargs_from_args()
        info_str = [f'({type(self).__name__} Info):',
                    f'Step Name: {self.name}',
                    f'Function Name: {self._func_name}',
                    f'Arguments Passed:',
                    f'{args_to_kwargs_dict if args_to_kwargs_dict else "N/A"}',
                    'Keyword-Arguments Set:',
                    f'{self.kwargs if self.kwargs else "N/A"}',
                    'Default Arguments:',
                    f'{self.get_function_args_not_set_in_kwargs()}']
        return '\n'.join(info_str)
    
    def __repr__(self):
        """
        Returns an unambiguous string representation of the FunctionBasedStep instance.

        Returns:
            str: A string representation showing how the FunctionBasedStep can be recreated.
        """
        cls_name = type(self).__name__
        full_func_name = f'{self.function.__module__}.{self._func_name}'
        info_str = [f'{cls_name}(', f'name={repr(self.name)},', f'function={full_func_name},']
        
        init_params = inspect.signature(self.__init__).parameters
        for arg_name in list(init_params)[2:-2]:
            info_str.append(f'{arg_name}={repr(getattr(self, arg_name))},')
        
        for arg_name, arg_val in zip(list(self.func_sig.parameters), self.args):
            info_str.append(f'{arg_name}={repr(arg_val)}', )
        
        for arg_name, arg_val in self.kwargs.items():
            info_str.append(f'{arg_name}={repr(arg_val)},')
        
        info_str.append(')')
        
        return f'\n    '.join(info_str)
    
    def all_args_non_empty_strings(self):
        """
        Checks if all positional arguments are non-empty strings.

        Returns:
            bool: True if all positional arguments are non-empty strings, False otherwise.
        """
        for arg in self.args:
            if arg == '':
                return False
        return True
    
    def all_kwargs_non_empty_strings(self):
        """
        Checks if all keyword arguments are non-empty strings.

        Returns:
            bool: True if all keyword arguments are non-empty strings, False otherwise.
        """
        for arg_name, arg in self.kwargs.items():
            if arg == '':
                return False
        return True
    
    def can_potentially_run(self):
        """
        Determines if the step can potentially be executed based on argument validation.
        Very simply checks if all arguments and keyword arguments are non-empty strings.

        Returns:
            bool: True if the step can potentially run, False otherwise.
        """
        return self.all_args_non_empty_strings() and self.all_kwargs_non_empty_strings()


class ObjectBasedStep(StepsAPI):
    """
    A step in a processing pipeline that is based on instantiating and invoking methods on an object.

    This class allows for initialization and execution of a specified object with given arguments and keyword arguments,
    validating that all mandatory parameters are provided.

    Attributes:
        name (str): The name of the step.
        class_type (type): The class type to be instantiated in this step.
        init_kwargs (ArgsDict): Keyword arguments for initializing the class.
        call_kwargs (ArgsDict): Keyword arguments for invoking the class.
        init_sig (inspect.Signature): The initialization signature of the class for validating arguments.
        call_sig (inspect.Signature): The call signature of the class for validating arguments.

    """
    def __init__(self, name: str, class_type: type, init_kwargs: dict, call_kwargs: dict) -> None:
        """
        Initializes an object-based step in the processing pipeline.

        Args:
            name (str): The name of the step.
            class_type (type): The class type to be instantiated in this step.
            init_kwargs (dict): Keyword arguments for initializing the class.
            call_kwargs (dict): Keyword arguments for invoking the class.
        """
        StepsAPI.__init__(self, skip_step=False)
        self.name: str = name
        self.class_type: type = class_type
        self.init_kwargs: ArgsDict = ArgsDict(init_kwargs)
        self.call_kwargs: ArgsDict = ArgsDict(call_kwargs)
        self.init_sig: inspect.Signature = inspect.signature(self.class_type.__init__)
        self.call_sig: inspect.Signature = inspect.signature(self.class_type.__call__)
        self.validate_kwargs()
    
    def validate_kwargs(self):
        """
        Validates that all mandatory initialization and call arguments have been provided.

        Raises:
            RuntimeError: If any mandatory arguments are missing.
        """
        empty_init_kwargs = self.get_empty_default_kwargs(self.init_sig, self.init_kwargs)
        empty_call_kwargs = self.get_empty_default_kwargs(self.call_sig, self.call_kwargs)
        
        if empty_init_kwargs or empty_call_kwargs:
            err_msg = [f"For {self.class_type.__name__}, the following arguments must be set:"]
            if empty_init_kwargs:
                err_msg.append("Initialization:")
                err_msg.append(f"{empty_init_kwargs}")
            if empty_call_kwargs:
                err_msg.append("Calling:")
                err_msg.append(f"{empty_call_kwargs}")
            raise RuntimeError("\n".join(err_msg))
    
    @staticmethod
    def get_args_not_set_in_kwargs(sig: inspect.Signature, kwargs: dict) -> dict:
        """
        Retrieves arguments of the signature that are not set in the keyword arguments.

        Args:
            sig (inspect.Signature): The signature of the function or method.
            kwargs (dict): The keyword arguments provided.

        Returns:
            dict: A dictionary of arguments that are not set in the keyword arguments.
        """
        unset_args_dict = ArgsDict()
        for arg_name, arg_val in sig.parameters.items():
            if arg_name not in kwargs and arg_name != 'self':
                if arg_val.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                    unset_args_dict[arg_name] = arg_val.default
        return unset_args_dict
    
    def get_empty_default_kwargs(self, sig: inspect.Signature, set_kwargs: dict) -> list:
        """
        Identifies arguments that have not been provided and lack default values.

        Args:
            sig (inspect.Signature): The signature of the function or method.
            set_kwargs (dict): The keyword arguments provided.

        Returns:
            list: A list of argument names that have no default values and are not provided.
        """
        unset_kwargs = self.get_args_not_set_in_kwargs(sig=sig, kwargs=set_kwargs)
        empty_kwargs = []
        for arg_name, arg_val in unset_kwargs.items():
            if arg_val is inspect.Parameter.empty:
                if arg_name not in set_kwargs:
                    empty_kwargs.append(arg_name)
        return empty_kwargs
    
    def execute(self) -> None:
        """
        Instantiates the class and invokes it with the provided arguments.

        Raises:
            The function may raise any exceptions that its implementation can throw.
        """
        print(f"(Info): Executing {self.name}")
        obj_instance = self.class_type(**self.init_kwargs)
        obj_instance(**self.call_kwargs)
        print(f"(Info): Finished {self.name}")
    
    def __str__(self):
        """
        Returns a detailed string representation of the ObjectBasedStep instance.

        Returns:
            str: A string describing the step, including its name, class, initialization, and call arguments.
        """
        unset_init_args = self.get_args_not_set_in_kwargs(self.init_sig, self.init_kwargs)
        unset_call_args = self.get_args_not_set_in_kwargs(self.call_sig, self.call_kwargs)
        
        info_str = [f'({type(self).__name__} Info):', f'Step Name: {self.name}',
                    f'Class Name: {self.class_type.__name__}', 'Initialization Arguments:', f'{self.init_kwargs}',
                    'Default Initialization Arguments:', f'{unset_init_args if unset_init_args else "N/A"}',
                    'Call Arguments:', f'{self.call_kwargs if self.call_kwargs else "N/A"}', 'Default Call Arguments:',
                    f'{unset_call_args if unset_call_args else "N/A"}']
        return '\n'.join(info_str)
    
    def __repr__(self):
        """
        Returns an unambiguous string representation of the ObjectBasedStep instance.

        Returns:
            str: A string representation showing how the ObjectBasedStep can be recreated.
        """
        cls_name = type(self).__name__
        full_func_name = f'{self.class_type.__module__}.{self.class_type.__name__}'
        info_str = [f'{cls_name}(', f'name={repr(self.name)},', f'class_type={full_func_name},']
        
        if self.init_kwargs:
            info_str.append('init_kwargs={')
            for arg_name, arg_val in self.init_kwargs.items():
                info_str.append(f'    {arg_name}={repr(arg_val)},')
            info_str[-1] = f'{info_str[-1]}' + '}'
        
        if self.call_kwargs:
            info_str.append('call_kwargs={')
            for arg_name, arg_val in self.call_kwargs.items():
                info_str.append(f'    {arg_name}={repr(arg_val)},')
            info_str[-1] = f'{info_str[-1]}' + '}'
        
        info_str.append(')')
        
        return f'\n    '.join(info_str)
    
    def all_init_kwargs_non_empty_strings(self):
        """
        Checks if all initialization keyword arguments are non-empty strings.

        Returns:
            bool: True if all initialization keyword arguments are non-empty strings, False otherwise.
        """
        for arg_name, arg_val in self.init_kwargs.items():
            if arg_val == '':
                return False
        return True
    
    def all_call_kwargs_non_empty_strings(self):
        """
        Checks if all call keyword arguments are non-empty strings.

        Returns:
            bool: True if all call keyword arguments are non-empty strings, False otherwise.
        """
        for arg_name, arg_val in self.call_kwargs.items():
            if arg_val == '':
                return False
        return True
    
    def can_potentially_run(self):
        """
        Determines if the step can potentially be executed based on argument validation.
        Very simply checks if all __init__ and __call__  keyword arguments for the object are
        non-empty strings.
        
        Returns:
            bool: True if the step can potentially run, False otherwise.
        """
        return self.all_init_kwargs_non_empty_strings() and self.all_call_kwargs_non_empty_strings()




"""
This module provides functions and a key class, :class:`GraphicalAnalysis`, for performing
graphical analysis on Time Activity Curve (TAC) data. It heavily utilizes Numpy and supports
various analysis methods like Patlak, Logan, and alternative Logan analysis.

The :class:`GraphicalAnalysis` class encapsulates the main functionality of the module. It provides
an organized way to perform graphical analysis where it initializes with paths to input data and
output details, runs an analysis using a specific method, calculates the best fit parameters,
computes properties related to the fitting process, and stores the analysis results. All these
properties and results are stored within an instance's 'analysis_props' dictionary, providing an
organized storage of result of an analysis task.

TODO:
    * Check if it makes more sense to lift out the more mathy methods out into a separate module.
    * Add references for the TCMs and Patlak. Could maybe rely on Turku PET Center.
    * Handle cases when tac_vals = 0.0. Might be able to use t_thresh so that we are past the 0-values.
    * Add more detailed documentation to patlak analysis.
    
"""

from collections.abc import Callable
from typing import Tuple
import warnings
import os
import json
import numba
import numpy as np
import pandas as pd
from ..utils.time_activity_curve import MultiTACAnalysisMixin, safe_load_tac
from ..utils.image_io import flatten_metadata


@numba.njit()
def _line_fitting_make_rhs_matrix_from_xdata(xdata: np.ndarray) -> np.ndarray:
    """Generates the RHS matrix for linear least squares fitting

    Args:
        xdata (np.ndarray): array of independent variable values

    Returns:
        np.ndarray: 2D matrix where first column is `xdata` and the second column is 1s.
        
    """
    out_matrix = np.ones((len(xdata), 2), float)
    out_matrix[:, 0] = xdata
    return out_matrix


@numba.njit()
def fit_line_to_data_using_lls(xdata: np.ndarray, ydata: np.ndarray) -> np.ndarray:
    """Find the linear least squares solution given the x and y variables.
    
    Performs a linear least squares fit to the provided data. Explicitly calls numpy's
    ``linalg.lstsq`` method by constructing the matrix equations. We assume that ``xdata`` and
    ``ydata`` have the same number of elements.
    
    Args:
        xdata: Array of independent variable values
        ydata: Array of dependent variable values

    Returns:
       Linear least squares solution. (m, b) values
       
    """
    make_2d_matrix = _line_fitting_make_rhs_matrix_from_xdata
    matrix = make_2d_matrix(xdata)
    fit_ans = np.linalg.lstsq(matrix, ydata)[0]
    return fit_ans


@numba.njit()
def fit_line_to_data_using_lls_with_rsquared(xdata: np.ndarray,
                                             ydata: np.ndarray) -> Tuple[float, float, float]:
    """Fits a line to the data using least squares and explicitly computes the r-squared value.

    Args:
        xdata (np.ndarray): X-coordinates of the data.
        ydata (np.ndarray): Y-coordinates of the data.

    Returns:
        tuple: A tuple containing three float values: the intercept of the fitted line, the slope
        of the fitted line, and the r-squared value.
    """
    make_2d_matrix = _line_fitting_make_rhs_matrix_from_xdata
    matrix = make_2d_matrix(xdata)
    fit_ans = np.linalg.lstsq(matrix, ydata)
    
    ss_res = fit_ans[1][0]
    ss_tot = np.sum((np.mean(ydata) - ydata) ** 2.)
    r_squared = 1.0 - ss_res / ss_tot
    return fit_ans[0][0], fit_ans[0][1], r_squared


@numba.njit()
def linear_least_squares_fit_with_stats(xdata: np.ndarray,
                                        ydata: np.ndarray) -> tuple[float, float, float, float, float]:
    """Fits a line to the data using least squares and explicitly computes:
        - Fit R^2
        - Standard error of the intercept
        - Standard error of the slope

    Args:
        xdata (np.ndarray): X-coordinates of the data.
        ydata (np.ndarray): Y-coordinates of the data.

    Returns:
        tuple: A tuple containing five float values: the intercept of the fitted line, the slope
        of the fitted line, the r-squared value, the intercept standard error, and the slope
        standard error.

    See:
        - https://mathworld.wolfram.com/LeastSquaresFitting.html
    """
    make_2d_matrix = _line_fitting_make_rhs_matrix_from_xdata
    matrix = make_2d_matrix(xdata)
    fit_ans = np.linalg.lstsq(matrix, ydata)

    x_mean = np.mean(xdata)
    y_mean = np.mean(ydata)

    ss_res = fit_ans[1][0]
    ss_tot = np.sum((y_mean - ydata) ** 2.)

    n = len(xdata)

    sum_square_xdiff = np.sum(xdata**2)-n*x_mean**2
    sum_square_ydiff = np.sum(ydata**2)-n*y_mean**2
    sum_square_xydiff = np.sum(xdata*ydata)-n*y_mean*x_mean

    s = np.sqrt((sum_square_ydiff-sum_square_xydiff**2/sum_square_xdiff)/(n-2))

    r_squared = 1.0 - ss_res / ss_tot

    se_intercept = s*np.sqrt(1/n+np.mean(xdata)**2/sum_square_xdiff)
    se_slope = s*sum_square_xdiff**(-0.5)
    return fit_ans[0][0], fit_ans[0][1], r_squared, se_intercept, se_slope

@numba.njit()
def cumulative_trapezoidal_integral(xdata: np.ndarray,
                                    ydata: np.ndarray,
                                    initial: float = 0.0) -> np.ndarray:
    """Calculates the cumulative integral of `ydata` over `xdata` using the trapezoidal rule.
    
    This function is based `heavily` on the :py:func:`scipy.integrate.cumulative_trapezoid`
    implementation (`source <https://github.com/scipy/scipy/blob/v1.12.0/scipy/integrate/_quadrature.py#L432-L536>`_).
    This implementation only works for 1D arrays and was implemented to work with :mod:`numba`.
    
    Args:
        xdata (np.ndarray): Array for the integration coordinate.
        ydata (np.ndarray): Array for the values to integrate

    Returns:
        (np.ndarray): Cumulative integral of ``ydata`` over ``xdata`` using the trapezoidal rule.
    """
    dx = np.diff(xdata)
    cum_int = np.zeros(len(xdata))
    cum_int[0] = initial
    cum_int[1:] = np.cumsum(dx * (ydata[1:] + ydata[:-1]) / 2.0)

    return cum_int


@numba.njit()
def calculate_patlak_x(tac_times: np.ndarray, tac_vals: np.ndarray) -> np.ndarray:
    r"""Calculates the x-variable in Patlak analysis
    :math:`\left(\frac{\int_{0}^{T}f(t)\mathrm{d}t}{f(T)}\right)`.
    
    Patlak-Gjedde analysis is a linearization of the 2-TCM with irreversible uptake in the second
    compartment. Therefore, we essentially have to fit a line to some data :math:`y = mx+b`. This
    function calculates the :math:`x` variable for Patlak analysis where,
    
    .. math::
    
        x = \frac{\int_{0}^{T} C_\mathrm{P}(t) \mathrm{d}t}{C_\mathrm{P}(t)},
    
    where further :math:`C_\mathrm{P}` is usually the plasma TAC.
    
    Args:
        tac_times (np.ndarray): Array containing the sampled times.
        tac_vals (np.ndarray): Array for activity values at the sampled times.

    Returns:
        (np.ndarray): Patlak x-variable. Cumulative integral of activity divided by activity.
        
    """
    cumulative_integral = cumulative_trapezoidal_integral(xdata=tac_times,
                                                          ydata=tac_vals,
                                                          initial=0.0)

    return cumulative_integral / tac_vals


@numba.njit()
def get_index_from_threshold(times_in_minutes: np.ndarray, t_thresh_in_minutes: float) -> int:
    """Get the index after which all times are greater than the threshold.

    Args:
        times_in_minutes (np.ndarray): Array containing times in minutes.
        t_thresh_in_minutes (float): Threshold time in minutes.

    Returns:
        int: Index for first time greater than or equal to the threshold time.
        
    Notes:
        If the threshold value is larger than the array, we return -1.
        
    """
    if t_thresh_in_minutes > np.max(times_in_minutes):
        return -1
    else:
        return np.argwhere(times_in_minutes >= t_thresh_in_minutes)[0, 0]


@numba.njit()
def patlak_analysis(tac_times_in_minutes: np.ndarray,
                    input_tac_values: np.ndarray,
                    region_tac_values: np.ndarray,
                    t_thresh_in_minutes: float) -> np.ndarray:
    """
    Performs Patlak analysis given the input TAC, region TAC, times and threshold.
    
    Args:
        tac_times_in_minutes (np.ndarray): Array of times in minutes.
        input_tac_values (np.ndarray): Array of input TAC values
        region_tac_values (np.ndarray): Array of ROI TAC values
        t_thresh_in_minutes (np.ndarray): Threshold time in minutes. Line is fit for all values
            after the threshold.

    Returns:
        (np.ndarray): Array containing :math:`(K_{i}, V_{0})` values.
    
    .. important::
        * We assume that the input TAC and ROI TAC values are sampled at the same times.
    
    """
    non_zero_indices = np.argwhere(input_tac_values != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.asarray([np.nan, np.nan])

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_minutes)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.asarray([np.nan, np.nan])

    patlak_x = calculate_patlak_x(tac_times=tac_times_in_minutes[non_zero_indices],
                                  tac_vals=input_tac_values[non_zero_indices])
    patlak_y = region_tac_values[non_zero_indices] / input_tac_values[non_zero_indices]

    patlak_values = fit_line_to_data_using_lls(xdata=patlak_x[t_thresh:], ydata=patlak_y[t_thresh:])

    return patlak_values


@numba.njit()
def patlak_analysis_with_rsquared(tac_times_in_minutes: np.ndarray,
                                  input_tac_values: np.ndarray,
                                  region_tac_values: np.ndarray,
                                  t_thresh_in_minutes: float) -> Tuple[float, float, float]:
    """Performs Patlak analysis given the input TAC, region TAC, times and threshold.

    Args:
        tac_times_in_minutes (np.ndarray): Array of times in minutes.
        input_tac_values (np.ndarray): Array of input TAC values
        region_tac_values (np.ndarray): Array of ROI TAC values
        t_thresh_in_minutes (np.ndarray): Threshold time in minutes. Line is fit for all values after the threshold.

    Returns:
        tuple: (slope, intercept, :math:`R^2`)

    .. important::
        * We assume that the input TAC and ROI TAC values are sampled at the same times.

    """
    non_zero_indices = np.argwhere(input_tac_values != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.nan, np.nan, np.nan

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_minutes)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.nan, np.nan, np.nan

    patlak_x = calculate_patlak_x(tac_times=tac_times_in_minutes[non_zero_indices],
                                  tac_vals=input_tac_values[non_zero_indices])
    patlak_y = region_tac_values[non_zero_indices] / input_tac_values[non_zero_indices]

    patlak_values = fit_line_to_data_using_lls_with_rsquared(xdata=patlak_x[t_thresh:], ydata=patlak_y[t_thresh:])

    return patlak_values


@numba.njit
def logan_analysis(tac_times_in_minutes: np.ndarray,
                   input_tac_values: np.ndarray,
                   region_tac_values: np.ndarray,
                   t_thresh_in_minutes: float) -> np.ndarray:
    """Performs Logan analysis on given input TAC, regional TAC, times and threshold, considering
    non-zero values.

    This function is similar to :func:`logan_analysis_with_rsquared`, but avoids the issue of
    division by zero by only considering non-zero TAC values for the region TAC since it is in
    the denominator. If the number of non-zero indices is less than or equal to 2, or if the number
    of time points after the threshold is less than or equal to 2, the function returns an array of
    NaNs.

    Args:
        tac_times_in_minutes (np.ndarray): Array of times in minutes.
        input_tac_values (np.ndarray): Array of input TAC values.
        region_tac_values (np.ndarray): Array of ROI TAC values.
        t_thresh_in_minutes (np.ndarray): Threshold time in minutes. Line is fit for all values
            after the threshold.

    Returns:
        np.ndarray: Array of two elements - (slope, intercept) of the best-fit line to the given
            data.

    .. important::
        * The interpretation of the returned values depends on the underlying kinetic model.
        * We assume that the input TAC and ROI TAC values are sampled at the same times.
        
    """
    non_zero_indices = np.argwhere(region_tac_values != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.asarray([np.nan, np.nan])

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_minutes)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.asarray([np.nan, np.nan])

    logan_x = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes, ydata=input_tac_values)
    logan_y = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes, ydata=region_tac_values)

    logan_x = logan_x[non_zero_indices][t_thresh:] / region_tac_values[non_zero_indices][t_thresh:]
    logan_y = logan_y[non_zero_indices][t_thresh:] / region_tac_values[non_zero_indices][t_thresh:]

    fit_ans = fit_line_to_data_using_lls(xdata=logan_x, ydata=logan_y)
    return fit_ans


@numba.njit()
def logan_analysis_with_rsquared(tac_times_in_minutes: np.ndarray,
                                 input_tac_values: np.ndarray,
                                 region_tac_values: np.ndarray,
                                 t_thresh_in_minutes: float) -> tuple[float, float, float]:
    """
    Performs Logan analysis on given input TAC, regional TAC, times and threshold.

    Args:
        tac_times_in_minutes (np.ndarray): Array of times in minutes.
        input_tac_values (np.ndarray): Array of input TAC values
        region_tac_values (np.ndarray): Array of ROI TAC values
        t_thresh_in_minutes (np.ndarray): Threshold time in minutes. Line is fit for all values after the threshold.

    Returns:
        tuple: (slope, intercept, :math:`R^2`)

    .. important::
        * The interpretation of the values depends on the underlying kinetic model.
        * We assume that the input TAC and ROI TAC values are sampled at the same times.
        
    """

    non_zero_indices = np.argwhere(region_tac_values != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.nan, np.nan, np.nan

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_minutes)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.nan, np.nan, np.nan

    logan_x = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes, ydata=input_tac_values)
    logan_y = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes, ydata=region_tac_values)

    logan_x = logan_x[non_zero_indices][t_thresh:] / region_tac_values[non_zero_indices][t_thresh:]
    logan_y = logan_y[non_zero_indices][t_thresh:] / region_tac_values[non_zero_indices][t_thresh:]

    logan_values = fit_line_to_data_using_lls_with_rsquared(xdata=logan_x, ydata=logan_y)

    return logan_values


@numba.njit()
def logan_ref_region_analysis(tac_times_in_minutes: np.ndarray,
                              input_tac_values: np.ndarray,
                              region_tac_values: np.ndarray,
                              t_thresh_in_minutes: float,
                              k2_prime: float) -> np.ndarray:
    """
    Performs Logan with reference region input function on given input TAC, regional TAC, times,
    threshold, and population averaged reference region k2.

    Args:
        tac_times_in_minutes (np.ndarray): Array of times in minutes.
        input_tac_values (np.ndarray): Array of input TAC values
        region_tac_values (np.ndarray): Array of ROI TAC values
        t_thresh_in_minutes (np.ndarray): Threshold time in minutes. Line is fit for all values
            after the threshold.
        k2_prime (float): Population averaged k2 value for the reference region.

    Returns:
        np.ndarray: Array of two elements - (slope, intercept) of the best-fit line to the given
            data.

    .. important::
        * The interpretation of the values depends on the underlying kinetic model.
        * We assume that the input TAC and ROI TAC values are sampled at the same times.
        
    """

    non_zero_indices = np.argwhere(region_tac_values != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.asarray([np.nan, np.nan])

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_minutes)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.asarray([np.nan, np.nan])

    logan_x = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes, ydata=input_tac_values)
    logan_y = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes, ydata=region_tac_values)

    logan_x_ref_region_term = input_tac_values[non_zero_indices][t_thresh:]/k2_prime
    logan_x_numerator = logan_x[non_zero_indices][t_thresh:] + logan_x_ref_region_term
    logan_x = logan_x_numerator / region_tac_values[non_zero_indices][t_thresh:]
    logan_y = logan_y[non_zero_indices][t_thresh:] / region_tac_values[non_zero_indices][t_thresh:]

    logan_values = fit_line_to_data_using_lls(xdata=logan_x, ydata=logan_y)

    return logan_values


@numba.njit()
def logan_ref_region_analysis_with_rsquared(tac_times_in_minutes: np.ndarray,
                                            input_tac_values: np.ndarray,
                                            region_tac_values: np.ndarray,
                                            t_thresh_in_minutes: float,
                                            k2_prime: float) -> tuple[float, float, float]:
    """
    Performs Logan with reference region input function on given input TAC, regional TAC, times,
    threshold, and population averaged reference region k2.

    Args:
        tac_times_in_minutes (np.ndarray): Array of times in minutes.
        input_tac_values (np.ndarray): Array of input TAC values
        region_tac_values (np.ndarray): Array of ROI TAC values
        t_thresh_in_minutes (np.ndarray): Threshold time in minutes. Line is fit for all values
            after the threshold.
        k2_prime (float): Population averaged k2 value for the reference region.

    Returns:
        tuple: (slope, intercept, :math:`R^2`)

    .. important::
        * The interpretation of the values depends on the underlying kinetic model.
        * We assume that the input TAC and ROI TAC values are sampled at the same times.
        
    """

    non_zero_indices = np.argwhere(region_tac_values != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.nan, np.nan, np.nan

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_minutes)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.nan, np.nan, np.nan

    logan_x = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes, ydata=input_tac_values)
    logan_y = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes, ydata=region_tac_values)

    logan_x_ref_region_term = input_tac_values[non_zero_indices][t_thresh:]/k2_prime
    logan_x_numerator = logan_x[non_zero_indices][t_thresh:] + logan_x_ref_region_term
    logan_x = logan_x_numerator / region_tac_values[non_zero_indices][t_thresh:]
    logan_y = logan_y[non_zero_indices][t_thresh:] / region_tac_values[non_zero_indices][t_thresh:]

    logan_values = fit_line_to_data_using_lls_with_rsquared(xdata=logan_x, ydata=logan_y)

    return logan_values


@numba.njit
def alternative_logan_analysis(tac_times_in_minutes: np.ndarray,
                               input_tac_values: np.ndarray,
                               region_tac_values: np.ndarray,
                               t_thresh_in_minutes: float) -> np.ndarray:
    """
    Performs Alternative Logan analysis on given input TAC, regional TAC, times and threshold,
    considering non-zero values for regional TAC.

    This function is similar to :func:`alternative_logan_analysis_with_rsquared`, but avoids the
    issue of division by zero by only considering non-zero TAC values for the region TAC since it
    is in the denominator. If the number of indices is less than or equal to 2, or if the number of
    time points after the threshold is less than or equal to 2, the function returns an array of
    NaNs.

    Args:
        tac_times_in_minutes (np.ndarray): Array of times in minutes.
        input_tac_values (np.ndarray): Array of input TAC values.
        region_tac_values (np.ndarray): Array of ROI TAC values.
        t_thresh_in_minutes (np.ndarray): Threshold time in minutes. Line is fit for all values
            after the threshold.

    Returns:
        np.ndarray: Array of two elements - (slope, intercept) of the best-fit line to the given
            data.

    .. important::
        * The interpretation of the returned values depends on the underlying kinetic model.
        * We assume that the input TAC and ROI TAC values are sampled at the same times.
        
    """
    non_zero_indices = np.argwhere(input_tac_values != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.asarray([np.nan, np.nan])

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_minutes)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.asarray([np.nan, np.nan])

    alt_logan_x = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes,
                                                  ydata=input_tac_values)
    alt_logan_y = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes,
                                                  ydata=region_tac_values)

    alt_logan_x = alt_logan_x[non_zero_indices][t_thresh:] / input_tac_values[non_zero_indices][t_thresh:]
    alt_logan_y = alt_logan_y[non_zero_indices][t_thresh:] / input_tac_values[non_zero_indices][t_thresh:]

    fit_ans = fit_line_to_data_using_lls(xdata=alt_logan_x, ydata=alt_logan_y)
    return fit_ans


@numba.njit()
def alternative_logan_analysis_with_rsquared(tac_times_in_minutes: np.ndarray,
                                             input_tac_values: np.ndarray,
                                             region_tac_values: np.ndarray,
                                             t_thresh_in_minutes: float) -> tuple[float, float, float]:
    """Performs alternative logan analysis on given input TAC, regional TAC, times and threshold.

    Args:
        tac_times_in_minutes (np.ndarray): Array of times in minutes.
        input_tac_values (np.ndarray): Array of input TAC values
        region_tac_values (np.ndarray): Array of ROI TAC values
        t_thresh_in_minutes (np.ndarray): Threshold time in minutes. Line is fit for all values
            after the threshold.

    Returns:
        tuple: (slope, intercept, :math:`R^2`)

    .. important::
        * The interpretation of the values depends on the underlying kinetic model.
        * We assume that the input TAC and ROI TAC values are sampled at the same times.
        
    """

    non_zero_indices = np.argwhere(input_tac_values != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.nan, np.nan, np.nan

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_minutes)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.nan, np.nan, np.nan

    alt_logan_x = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes,
                                                  ydata=input_tac_values)
    alt_logan_y = cumulative_trapezoidal_integral(xdata=tac_times_in_minutes,
                                                  ydata=region_tac_values)

    alt_logan_x = alt_logan_x[non_zero_indices][t_thresh:] / input_tac_values[non_zero_indices][t_thresh:]
    alt_logan_y = alt_logan_y[non_zero_indices][t_thresh:] / input_tac_values[non_zero_indices][t_thresh:]

    alt_logan_values = fit_line_to_data_using_lls_with_rsquared(xdata=alt_logan_x,
                                                                ydata=alt_logan_y)

    return alt_logan_values


def get_graphical_analysis_method(method_name: str) -> Callable:
    """
    Function for obtaining the appropriate graphical analysis method.

    This function accepts a string specifying a graphical time-activity curve (TAC) analysis
    method. It returns a reference to the function that performs the selected analysis method.

    Args:
        method_name (str): The name of the graphical method. This should be one of the following
            strings: 'patlak', 'logan', 'logan_ref', or 'alt_logan'.

    Returns:
        function: A reference to the function that performs the corresponding graphical TAC
        analysis. The returned function will take arguments specific to the analysis method, such
        as input TAC values, tissue TAC values, TAC times in minutes, and threshold time in
        minutes.
    
    Raises:
        ValueError: If `method_name` is not one of the supported graphical analysis methods, i.e.,
            'patlak', 'logan', 'logan_ref', or 'alt_logan'.
    
    Example:
        .. code-block:: python
            
            from petpal.graphical_analysis import get_graphical_analysis_method as get_method
            from petpal.utils.image_io import safe_load_tac as load_tac
            
            selected_func = get_method('logan')
            input_tac_values, tac_times_in_minutes = load_tac("PATH/TO/PLASMA/TAC.tsv")
            tissue_tac_values, _ = load_tac("PATH/TO/ROI/TAC.tsv")
            
            results = selected_func(tac_times_in_minutes,
                                    input_tac_values,
                                    tissue_tac_values,
                                    t_thresh_in_minutes)
            print(results)
                                    
    """
    match method_name:
        case "patlak":
            return patlak_analysis
        case "alt_logan":
            return alternative_logan_analysis
        case "logan":
            return logan_analysis
        case "logan_ref":
            return logan_ref_region_analysis
        case _:
            raise ValueError("Invalid method_name! Must be either 'patlak', 'logan', 'alt_logan',"
                             f"'logan_ref'. Got {method_name}")


def get_graphical_analysis_method_with_rsquared(method_name: str) -> Callable:
    """
    Function for obtaining the appropriate graphical analysis method which also calculates the
    r-squared value.

    This function accepts a string specifying a graphical time-activity curve (TAC) analysis
    method. It returns a reference to the function that performs the selected analysis method.

    Args:
        method_name (str): The name of the graphical method. This should be one of the following
            strings: 'patlak', 'logan', 'logan_ref', or 'alt_logan'.

    Returns:
        function: A reference to the function that performs the corresponding graphical TAC
            analysis. The returned function will take arguments specific to the analysis method,
            such as input TAC values, tissue TAC values, TAC times in minutes, and threshold time
            in minutes.

    Raises:
        ValueError: If `method_name` is not one of the supported graphical analysis methods, i.e.,
            'patlak', 'logan', 'logan_ref' or 'alt_logan'.

    Example:
        .. code-block:: python
            
            from petpal.graphical_analysis import get_graphical_analysis_method_with_rsquared as get_method
            from petpal.graphical_analysis import safe_load_tac as load_tac
            
            selected_func = get_method('logan')
            input_tac_values, tac_times_in_minutes = load_tac("PATH/TO/PLASMA/TAC.tsv")
            tissue_tac_values, _ = load_tac("PATH/TO/ROI/TAC.tsv")
            
            results = selected_func(input_tac_values,
                                    tissue_tac_values,
                                    tac_times_in_minutes,
                                    t_thresh_in_minutes)
            print(results)
            
    """
    match method_name:
        case "patlak":
            return patlak_analysis_with_rsquared
        case "alt_logan":
            return alternative_logan_analysis_with_rsquared
        case "logan":
            return logan_analysis_with_rsquared
        case "logan_ref":
            return logan_ref_region_analysis_with_rsquared
        case _:
            raise ValueError(f"Invalid method_name! Must be either 'patlak', 'logan', 'alt_logan', 'logan_ref'. Got {method_name}")


def km_multifit_analysis_to_tsv(analysis_props: list[dict],
                                output_directory: str,
                                output_filename_prefix: str,
                                method: str,
                                inferred_seg_labels: list[str]):
    """
    Saves the analysis results to a TSV file as a table with fit parameters for each ROI.

    Args:
        analysis_props (list[dict]): List of fit results for each region.
        output_directory (str): Directory where results are saved.
        output_filename_prefix (str): Prefix for the output file.
        method (str): Name of the method for the model.
        inferred_seg_labels (list[str]): Names of each region used in the analysis.
    
    Raises:
        RuntimeError: If 'run_analysis' method has not been called before save_analysis.
    """
    filename = f'{output_filename_prefix}_desc-{method}_fitprops.tsv'
    filepath = os.path.join(output_directory, filename)
    fit_table = pd.DataFrame()
    for seg_name, fit_props in zip(inferred_seg_labels, analysis_props):
        tmp_table = pd.DataFrame(flatten_metadata(fit_props),index=[seg_name])
        fit_table = pd.concat([fit_table,tmp_table])
    fit_table.T.to_csv(filepath, sep='\t')


def km_multifit_analysis_to_jsons(analysis_props: list[dict],
                                  output_directory: str,
                                  output_filename_prefix: str,
                                  method: str,
                                  inferred_seg_labels: list[str]):
    """
    Saves the analysis results to a JSON file for each segment/TAC.

    Args:
        analysis_props (list[dict]): List of fit results for each region.
        output_directory (str): Directory where results are saved.
        output_filename_prefix (str): Prefix for the output files.
        method (str): Name of the method for the model.
        inferred_seg_labels (list[str]): Names of each region used in the analysis.

    Raises:
        RuntimeError: If 'run_analysis' method has not been called before 'save_analysis'.
    """
    for seg_name, fit_props in zip(inferred_seg_labels, analysis_props):
        filename = [output_filename_prefix,
                    f'desc-{method}',
                    f'seg-{seg_name}',
                    'fitprops.json']
        filename = '_'.join(filename)
        filepath = os.path.join(output_directory, filename)

        with open(filepath, 'w', encoding='utf-8') as f:
            json.dump(obj=fit_props, fp=f, indent=4)


class GraphicalAnalysis:
    """
    :class:`GraphicalAnalysis` to handle Graphical Analysis for time activity curve (TAC) data.

    The :class:`GraphicalAnalysis` class provides methods for managing TAC data analysis. The class
    is initialized with paths to input TAC data, region of interest (ROI) TAC data files, an output
    directory, and output filename prefix.

    Analysis is performed by specifying a method name and threshold time in the
    :func:`run_analysis` method. The results of the analysis are stored within the instance's
    'analysis_props' dictionary.

    Key methods include:
    - :func:`init_analysis_props`: Initializes a dictionary with keys for analysis properties and default values.
    - :func:`run_analysis`: Runs the graphical analysis on the data using a specific method.
    - :func:`calculate_fit`: Calculates the best fit values for a given graphical analysis method.
    - :func:`calculate_fit_properties`: Calculates and stores the properties related to the fitting process.
    - :func:`save_analysis`: Stores the 'analysis_props' dictionary into a JSON file.

    Raises:
        RuntimeError: If the :func:`run_analysis` method has not been run before
            :func:`save_analysis` method.

    Attributes:
        input_tac_path (str): The path to input TAC data file.
        roi_tac_path (str): The path to ROI TAC data file.
        output_directory (str): Directory where the output should be saved.
        output_filename_prefix (str): Output filename prefix for saving the analysis.
        analysis_props (dict): Property dictionary used to store results of the analysis.
        method (str): The name of the graphical analysis method to be utilised.
        fit_thresh_in_mins (float): The fitting threshold time in minutes for the analysis method.
        self.analysis_func (Callable): The function used for performing the fitting.
        
    """
    def __init__(self,
                 input_tac_path: str,
                 roi_tac_path: str,
                 output_directory: str,
                 output_filename_prefix: str,
                 method: str,
                 fit_thresh_in_mins: float) -> None:
        """
        Initializes GraphicalAnalysis with provided paths and output details.

        Args:
            input_tac_path (str): The path to the file containing input Time Activity Curve (TAC)
                data.
            roi_tac_path (str): The path to the file containing Region of Interest (ROI) TAC data.
            output_directory (str): The directory where the output of the analysis should be saved.
            output_filename_prefix (str): The prefix for the name of output file(s).
            method (str): The name of the graphical analysis method to be utilised.
            fit_thresh_in_mins (float): The fitting threshold time in minutes for the analysis
                method.
    
        Returns:
            None
        """
        self.input_tac_path = os.path.abspath(input_tac_path)
        self.roi_tac_path = os.path.abspath(roi_tac_path)
        self.output_directory = os.path.abspath(output_directory)
        self.output_filename_prefix = output_filename_prefix
        self.analysis_props = self.init_analysis_props()
        self.method = method
        self.fit_thresh_in_mins = fit_thresh_in_mins
        self.analysis_func = get_graphical_analysis_method_with_rsquared(method_name=self.method)

    def init_analysis_props(self) -> dict:
        """
        Initializes analysis properties dictionary.

        This method initializes a dictionary with keys for all the analysis properties and default
        values set to None. The paths to the input TAC and ROI TAC files are set from the object's
        properties.

        Returns:
            dict: A dictionary with keys for each analysis property and default values. The keys
                include 'FilePathPTAC' (the file path to the input TAC file), 'FilePathTTAC' (the
                file path to the ROI TAC file), 'MethodName' (the name of the method used in the
                analysis), 'ThresholdTime' (the threshold time for the analysis), 'StartFrameTime'
                and 'EndFrameTime' (the start and end times for the frame), 'NumberOfPointsFit'
                (The number of points used in the fit), 'Slope', 'Intercept' and 'RSquared' (the
                slope, intercept and R-squared of the fit).
        """
        props = {'FilePathPTAC': self.input_tac_path,
                 'FilePathTTAC': self.roi_tac_path,
                 'MethodName': None,
                 'ThresholdTime': None,
                 'StartFrameTime': None,
                 'EndFrameTime': None,
                 'NumberOfPointsFit': None,
                 'Slope': None,
                 'Intercept': None,
                 'RSquared': None}
        return props

    def run_analysis(self, **run_kwargs):
        """
        Runs the graphical analysis on the data using the specified method.

        This method is the main entry point to carry out the analysis. It executes the steps in
        order, first calculating the fit, then calculating the properties of the fit.

        Args:
            run_kwargs: Additional keyword arguments used in the analysis. These are passed on to
                :meth:`calculate_fit` and :meth:`calculate_fit_properties`.

        Returns:
            None

        Side Effects:
            Computes and updates the analysis-related properties in the object based on the
                provided method and threshold.
        """
        self.calculate_fit(**run_kwargs)
        self.calculate_fit_properties(**run_kwargs)

    def calculate_fit(self, **run_kwargs):
        """
        Calculates the best fit parameters for a graphical analysis method.
    
        This method applies the specified graphical analysis method to the Time Activity Curve
        (TAC) data and stores the slope, intercept, and r-squared values of the fit in the
        analysis properties.
    
        Args:
            run_kwargs: Additional keyword arguments used in the analysis. These are passed on to
                `analysis_func`.
    
        Returns:
            None
            
        Side Effect:
            Updates 'Slope', 'Intercept', and 'RSquared' in self.analysis_props dictionary with
                calculated fit parameters.

        """
        p_tac_times, p_tac_vals = safe_load_tac(self.input_tac_path)
        _t_tac_times, t_tac_vals = safe_load_tac(self.roi_tac_path)
        slope, intercept, rsquared = self.analysis_func(tac_times_in_minutes=p_tac_times,
                                                        input_tac_values=p_tac_vals,
                                                        region_tac_values=t_tac_vals,
                                                        t_thresh_in_minutes=self.fit_thresh_in_mins,
                                                        **run_kwargs)
        self.analysis_props['Slope'] = slope
        self.analysis_props['Intercept'] = intercept
        self.analysis_props['RSquared'] = rsquared

    def calculate_fit_properties(self, **run_kwargs):
        """
        Calculates and stores the properties related to the fitting process.

        This method calculates several properties related to the fitting process, including the
        threshold time, the name of the method used, the start and end frame time, and the number
        of points used in the fit. These values are stored in the instance's `analysis_props`
        variable.

        Args:
            run_kwargs: Additional keyword arguments used in the analysis. These are saved to the
                analysis properties as individual properties.


        Note:
            This method relies on the :func:`safe_load_tac` function to load time-activity curve
            (TAC) data from the file at ``input_tac_path``, and the
            :func:`get_index <get_index_from_threshold>` function to get the index from the
            threshold time.

        See also:
            * :func:`safe_load_tac`: Function to safely load TAC data from a file.
            * :func:`get_index_from_threshold`: Function to get the index from the threshold time.

        Returns:
            None. The results are stored within the instance's `analysis_props` variable.
        """
        self.analysis_props['ThresholdTime'] = self.fit_thresh_in_mins

        for analysis_parameter_key, analysis_parameter_val in run_kwargs.items():
            self.analysis_props[analysis_parameter_key] = analysis_parameter_val

        self.analysis_props['MethodName'] = self.method
        p_tac_times, _ = safe_load_tac(filename=self.input_tac_path)
        t_thresh_index = get_index_from_threshold(times_in_minutes=p_tac_times,
                                                  t_thresh_in_minutes=self.fit_thresh_in_mins)
        self.analysis_props['StartFrameTime'] = p_tac_times[t_thresh_index]
        self.analysis_props['EndFrameTime'] = p_tac_times[-1]
        self.analysis_props['NumberOfPointsFit'] = len(p_tac_times[t_thresh_index:])

    def save_analysis(self):
        """
        Saves the analysis properties to a JSON file.

        This method saves the 'analysis_props' dictionary to a JSON file. The file is saved in the
        specified output directory under a filename constructed from the 'output_filename_prefix'
        and the method name used in analysis. Before executing, the method checks if analysis has
        been run by verifying if 'RSquared' in 'analysis_props' is not None.
    
        Raises:
            RuntimeError: If the 'run_analysis' method was not called before 'save_analysis' is
                invoked.
    
        Returns:
            None
        """
        if self.analysis_props['RSquared'] is None:
            raise RuntimeError("'run_analysis' method must be called before 'save_analysis'.")
        out_img_prefix = f"{self.output_filename_prefix}_desc-{self.analysis_props['MethodName']}"
        file_dir_prefix = os.path.join(self.output_directory, out_img_prefix)
        analysis_props_file = f"{file_dir_prefix}_fitprops.json"
        with open(analysis_props_file, 'w',encoding='utf-8') as f:
            json.dump(obj=self.analysis_props, fp=f, indent=4)

    def __call__(self, **run_kwargs):
        """
        Runs :meth:`run_analysis` and :meth:`save_analysis` to run the analysis and save the
        analysis properties.
        
        Args:
            run_kwargs: Additional keyword arguments used in the analysis. These are passed on to
                :meth:`run_analysis`.
        """
        self.run_analysis(**run_kwargs)
        self.save_analysis()


class MultiTACGraphicalAnalysis(GraphicalAnalysis, MultiTACAnalysisMixin):
    """
    A class that performs graphical analysis on multiple tissue TACs (Time Activity Curves).

    Attributes:
        input_tac_path (str): Path to the input TAC file.
        roi_tacs_dir (str): Directory containing region of interest TAC files.
        output_directory (str): Directory for saving analysis results.
        output_filename_prefix (str): Prefix for output filenames.
        method (str): Method used for analysis.
        fit_thresh_in_mins (Optional[float]): Threshold in minutes for fit calculation.
    """
    def __init__(self,
                 input_tac_path: str,
                 roi_tacs_dir:str,
                 output_directory: str,
                 output_filename_prefix: str,
                 method:str,
                 fit_thresh_in_mins=None,):
        """
        Initializes the MultiTACGraphicalAnalysis object with required paths, method, and threshold.

        Args:
            input_tac_path (str): Path to the input TAC file.
            roi_tacs_dir (str): Directory containing region of interest TAC files.
            output_directory (str): Directory for saving analysis results.
            output_filename_prefix (str): Prefix for output filenames.
            method (str): Method used for analysis.
            fit_thresh_in_mins (Optional[float], optional): Threshold in minutes for fit
                calculation. Defaults to None.
        """
        MultiTACAnalysisMixin.__init__(self,
                                       input_tac_path=input_tac_path,
                                       tacs_dir=roi_tacs_dir)
        GraphicalAnalysis.__init__(self,
                                   input_tac_path=input_tac_path,
                                   roi_tac_path=roi_tacs_dir,
                                   output_directory=output_directory,
                                   output_filename_prefix=output_filename_prefix,
                                   method=method,
                                   fit_thresh_in_mins=fit_thresh_in_mins
                                   )

    def init_analysis_props(self):
        """
        Initializes analysis properties for each tissue TAC. Overrides
        :meth:`GraphicalAnalysis.init_analysis_props`.

        Returns:
            list[dict]: A list of analysis property dictionaries for each TAC.
        """
        num_of_tacs = self.num_of_tacs
        analysis_props = [GraphicalAnalysis.init_analysis_props(self) for a_tac in range(num_of_tacs)]
        for tac_id, a_prop_dict in enumerate(analysis_props):
            a_prop_dict['FilePathTTAC'] = os.path.abspath(self.tacs_files_list[tac_id])
        return analysis_props


    def calculate_fit(self, **run_kwargs):
        """
        Calculates the fit for each TAC, updating the analysis properties with slope, intercept,
        and R-squared values. Overrides :meth:`GraphicalAnalysis.calculate_fit`.

        Args:
            run_kwargs: Additional keyword arguments passed on to `analysis_func`.
        """
        p_tac_times, p_tac_vals = safe_load_tac(self.input_tac_path)
        for tac_id, a_tac in enumerate(self.tacs_files_list):
            _, t_tac_vals = safe_load_tac(a_tac)
            try:
                slope, intercept, rsquared = self.analysis_func(tac_times_in_minutes=p_tac_times,
                                                                input_tac_values=p_tac_vals,
                                                                region_tac_values=t_tac_vals,
                                                                t_thresh_in_minutes=self.fit_thresh_in_mins,
                                                                **run_kwargs)
            except np.linalg.LinAlgError:
                slope, intercept, rsquared = np.nan, np.nan, np.nan
            self.analysis_props[tac_id]['Slope'] = slope
            self.analysis_props[tac_id]['Intercept'] = intercept
            self.analysis_props[tac_id]['RSquared'] = rsquared

    def calculate_fit_properties(self, **run_kwargs):
        """
        Calculates additional properties of the fit, such as threshold, method name,
        start and end times, and number of points. Iterates over all the TACs.
        
        Overrides :meth:`GraphicalAnalysis.calculate_fit_properties`

        Args:
            run_kwargs: Additional keyword arguments used in the analysis. These are saved to the
                analysis properties as individual properties.

        """
        p_tac_times, _ = safe_load_tac(self.input_tac_path)
        t_thresh_index = get_index_from_threshold(times_in_minutes=p_tac_times,
                                                  t_thresh_in_minutes=self.fit_thresh_in_mins)
        points_fit = len(p_tac_times[t_thresh_index:])
        start_time=p_tac_times[t_thresh_index]
        end_time=p_tac_times[-1]

        for tac_id, _a_tac in enumerate(self.tacs_files_list):
            self.analysis_props[tac_id]['ThresholdTime'] = self.fit_thresh_in_mins
            for analysis_parameter_key, analysis_parameter_val in run_kwargs.items():
                self.analysis_props[tac_id][analysis_parameter_key] = analysis_parameter_val
            self.analysis_props[tac_id]['MethodName'] = self.method
            self.analysis_props[tac_id]['StartFrameTime'] = start_time
            self.analysis_props[tac_id]['EndFrameTime'] = end_time
            self.analysis_props[tac_id]['NumberOfPointsFit'] = points_fit

    def save_analysis(self, output_as_tsv: bool=True, output_as_json: bool=False):
        """
        Saves the analysis results to a TSV file as a table with fit parameters for each ROI.

        Args:
            output_as_tsv (bool): Set True to write results to TSV table. Default True.
            output_as_json (bool): Set True to write results to a folder with one JSON file per
                region. Default False.

        Raises:
            RuntimeError: If 'run_analysis' method has not been called before save_analysis.
        """
        if self.analysis_props[0]['RSquared'] is None:
            raise RuntimeError("'run_analysis' method must be called before 'save_analysis'.")

        if output_as_json:
            km_multifit_analysis_to_jsons(analysis_props=self.analysis_props,
                                          output_directory=self.output_directory,
                                          output_filename_prefix=self.output_filename_prefix,
                                          method=self.method,
                                          inferred_seg_labels=self.inferred_seg_labels)
        if output_as_tsv:
            km_multifit_analysis_to_tsv(analysis_props=self.analysis_props,
                                        output_directory=self.output_directory,
                                        output_filename_prefix=self.output_filename_prefix,
                                        method=self.method,
                                        inferred_seg_labels=self.inferred_seg_labels)
        if not output_as_tsv and not output_as_json:
            warnings.warn('Both output_as_tsv and output_as_json set False. Results not written to '
                          'file.')

    def __call__(self, output_as_tsv: bool=True, output_as_json: bool=False, **run_kwargs):
        """
        Runs :meth:`run_analysis` and :meth:`save_analysis` to run the analysis and save the
        analysis properties.
        
        Args:
            output_as_tsv (bool): Set True to write results to TSV table. Default True.
            output_as_json (bool): Set True to write results to a folder with one JSON file per
                region. Default False.
            run_kwargs: Additional keyword arguments used in the analysis. These are passed on to
                :meth:`run_analysis`.
        """
        self.run_analysis(**run_kwargs)
        self.save_analysis(output_as_tsv=output_as_tsv, output_as_json=output_as_json)
